__author__ = "sibirrer"

import pytest
import numpy.testing as npt
import numpy as np

from lenstronomy.LensModel.profile_integrals import ProfileIntegrals


class TestNumerics(object):
    """Tests the second derivatives of various lens models."""

    def setup_method(self):
        pass

    def assert_integrals(self, Model, kwargs):
        lensModel = Model()
        int_profile = ProfileIntegrals(lensModel)
        r = 2.0

        density2d_num = int_profile.density_2d(r, kwargs)
        density2d = lensModel.density_2d(r, 0, **kwargs)
        npt.assert_almost_equal(density2d / density2d_num, 1.0, decimal=1)

        kwargs["center_x"] = 0
        kwargs["center_y"] = 0
        mass_2d_num = int_profile.mass_enclosed_2d(r, kwargs)
        del kwargs["center_x"]
        del kwargs["center_y"]
        mass_2d = lensModel.mass_2d(r, **kwargs)
        npt.assert_almost_equal(mass_2d / mass_2d_num, 1, decimal=1)

        kwargs["center_x"] = 0
        kwargs["center_y"] = 0
        mass_3d_num = int_profile.mass_enclosed_3d(r, kwargs)
        del kwargs["center_x"]
        del kwargs["center_y"]
        mass_3d = lensModel.mass_3d(r, **kwargs)
        npt.assert_almost_equal(mass_3d / mass_3d_num, 1, decimal=2)

    def assert_lens_integrals(self, Model, kwargs, pi_convention=True):
        """Checks whether the integral in projection of the density_lens() function is
        the convergence.

        :param Model: lens model instance
        :param kwargs: keyword arguments of lens model
        :return:
        """
        lensModel = Model()
        int_profile = ProfileIntegrals(lensModel)
        r = 2.0
        kappa_num = int_profile.density_2d(r, kwargs, lens_param=True)
        f_xx, f_xy, f_yx, f_yy = lensModel.hessian(r, 0, **kwargs)
        kappa = 1.0 / 2 * (f_xx + f_yy)
        npt.assert_almost_equal(kappa_num, kappa, decimal=2)
        try:
            del kwargs["center_x"]
            del kwargs["center_y"]
        except:
            pass
        bool_mass_2d_lens = False
        try:
            mass_2d = lensModel.mass_2d_lens(r, **kwargs)
            bool_mass_2d_lens = True
        except:
            pass
        if bool_mass_2d_lens:
            alpha_x, alpha_y = lensModel.derivatives(r, 0, **kwargs)
            alpha = np.sqrt(alpha_x**2 + alpha_y**2)
            if pi_convention:
                npt.assert_almost_equal(alpha, mass_2d / r / np.pi, decimal=5)
            else:
                npt.assert_almost_equal(alpha, mass_2d / r, decimal=5)
        try:
            mass_3d = lensModel.mass_3d_lens(r, **kwargs)
            bool_mass_3d_lens = True
        except:
            bool_mass_3d_lens = False
        if bool_mass_3d_lens:
            mass_3d_num = int_profile.mass_enclosed_3d(
                r, kwargs_profile=kwargs, lens_param=True
            )
            print(mass_3d, mass_3d_num, "test num")
            npt.assert_almost_equal(mass_3d / mass_3d_num, 1, decimal=2)

    def test_PJaffe(self):
        kwargs = {"rho0": 1.0, "Ra": 0.2, "Rs": 2.0}
        from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe as Model

        self.assert_integrals(Model, kwargs)

    def test_PJaffa_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe as Model

        lensModel = Model()
        sigma0 = 1.0
        Ra = 0.2
        Rs = 2.0
        rho0 = lensModel.sigma2rho(sigma0, Ra, Rs)
        kwargs_lens = {
            "sigma0": sigma0,
            "Ra": Ra,
            "Rs": Rs,
            "center_x": 0,
            "center_y": 0,
        }
        kwargs_density = {"rho0": rho0, "Ra": Ra, "Rs": Rs}
        r = 1.0
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_nfw(self):
        kwargs = {"rho0": 1.0, "Rs": 5.0, "center_x": 0, "center_y": 0}
        from lenstronomy.LensModel.Profiles.nfw import NFW as Model

        self.assert_integrals(Model, kwargs)

        kwargs_lens = {"alpha_Rs": 1.0, "Rs": 5.0, "center_x": 0, "center_y": 0}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_nfw_ellipse(self):
        kwargs = {
            "rho0": 1.0,
            "Rs": 5.0,
            "center_x": 0,
            "center_y": 0,
            "e1": 0,
            "e2": 0,
        }
        from lenstronomy.LensModel.Profiles.nfw_ellipse import NFW_ELLIPSE as Model

        # self.assert_integrals(Model, kwargs)

        kwargs_lens = {
            "alpha_Rs": 1.0,
            "Rs": 5.0,
            "center_x": 0,
            "center_y": 0,
            "e1": 0,
            "e2": 0,
        }
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_nfw_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.nfw import NFW as Model

        lensModel = Model()
        alpha_Rs = 1.0
        Rs = 2.0
        rho0 = lensModel.alpha2rho0(alpha_Rs, Rs)
        kwargs_lens = {"alpha_Rs": alpha_Rs, "Rs": Rs}
        kwargs_density = {"rho0": rho0, "Rs": Rs}
        self.assert_lens_integrals(Model, kwargs_lens)
        self.assert_integrals(Model, kwargs_density)
        r = 2.0
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_hernquist(self):
        from lenstronomy.LensModel.Profiles.hernquist import Hernquist as Model

        kwargs = {"rho0": 1.0, "Rs": 5.0}
        self.assert_integrals(Model, kwargs)
        kwargs = {"sigma0": 1.0, "Rs": 5.0}
        self.assert_lens_integrals(Model, kwargs)

    def test_hernquist_ellipse(self):
        from lenstronomy.LensModel.Profiles.hernquist_ellipse import (
            Hernquist_Ellipse as Model,
        )

        kwargs = {"rho0": 1.0, "Rs": 5.0, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)
        kwargs = {"sigma0": 1.0, "Rs": 5.0, "e1": 0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs)

    def test_hernquist_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.hernquist import Hernquist as Model

        lensModel = Model()
        sigma0 = 1.0
        Rs = 2.0
        rho0 = lensModel.sigma2rho(sigma0, Rs)
        kwargs_lens = {"sigma0": sigma0, "Rs": Rs}
        kwargs_density = {"rho0": rho0, "Rs": Rs}
        r = 0.5
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_sis(self):
        from lenstronomy.LensModel.Profiles.sis import SIS as Model

        kwargs = {"rho0": 1.0}
        self.assert_integrals(Model, kwargs)
        kwargs_lens = {"theta_E": 1.0}
        self.assert_lens_integrals(Model, kwargs_lens, pi_convention=True)

    def test_sis_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.sis import SIS as Model

        lensModel = Model()
        theta_E = 1.0
        rho0 = lensModel.theta2rho(theta_E)
        kwargs_lens = {"theta_E": theta_E}
        kwargs_density = {"rho0": rho0}
        r = 0.5
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)
        lensModel.density_2d(1, 1, rho0=1)

    def test_sie(self):
        from lenstronomy.LensModel.Profiles.sie import SIE as Model

        kwargs = {"rho0": 1.0, "e1": 0, "e2": 0}
        self.assert_integrals(Model, kwargs)
        kwargs_lens = {"theta_E": 1.0, "e1": 0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_spep(self):
        from lenstronomy.LensModel.Profiles.spep import SPEP as Model

        kwargs_lens = {"theta_E": 1, "gamma": 2, "e1": 0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_sie_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.sie import SIE as Model

        lensModel = Model()
        theta_E = 1.0
        rho0 = lensModel.theta2rho(theta_E)
        kwargs_lens = {"theta_E": theta_E, "e1": 0, "e2": 0}
        kwargs_density = {"rho0": rho0, "e1": 0, "e2": 0}
        r = 0.5
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_spp(self):
        from lenstronomy.LensModel.Profiles.spp import SPP as Model

        kwargs = {"rho0": 10.0, "gamma": 2.2}
        self.assert_integrals(Model, kwargs)

        kwargs = {"rho0": 1.0, "gamma": 2.0}
        self.assert_integrals(Model, kwargs)
        kwargs_lens = {"theta_E": 1.0, "gamma": 2.0}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_spp_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.spp import SPP as Model

        lensModel = Model()
        theta_E = 1.0
        gamma = 2.2
        rho0 = lensModel.theta2rho(theta_E, gamma)
        kwargs_lens = {"theta_E": theta_E, "gamma": gamma}
        kwargs_density = {"rho0": rho0, "gamma": gamma}
        r = 0.5
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_gaussian(self):
        from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa as Model

        kwargs = {"amp": 1.0 / 4.0, "sigma": 2.0}
        self.assert_integrals(Model, kwargs)

    def test_gaussian_density_deflection(self):
        """Tests whether the unit conversion between the lensing parameter 'sigma0' and
        the units in the density profile are ok :return:"""

        from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa as Model

        lensModel = Model()
        amp = 1.0 / 4.0
        sigma = 2.0
        amp_lens = lensModel._amp3d_to_2d(amp, sigma, sigma)
        kwargs_lens = {"amp": amp_lens, "sigma": sigma}
        kwargs_density = {"amp": amp, "sigma": sigma}
        r = 0.5
        mass_2d = lensModel.mass_2d(r, **kwargs_density)
        alpha_mass = mass_2d / r
        alpha_r, _ = lensModel.derivatives(r, 0, **kwargs_lens)
        npt.assert_almost_equal(alpha_mass / np.pi, alpha_r, decimal=5)

    def test_coreBurk(self):
        from lenstronomy.LensModel.Profiles.coreBurkert import CoreBurkert as Model

        kwargs = {"rho0": 1.0, "Rs": 10, "r_core": 5}
        self.assert_integrals(Model, kwargs)

        kwargs = {"rho0": 1.0, "Rs": 9, "r_core": 11}
        self.assert_integrals(Model, kwargs)

    def test_tnfw(self):
        from lenstronomy.LensModel.Profiles.tnfw import TNFW as Model

        kwargs = {"rho0": 1.0, "Rs": 1, "r_trunc": 4}
        self.assert_integrals(Model, kwargs)

    def test_cnfw(self):
        from lenstronomy.LensModel.Profiles.cnfw import CNFW as Model

        kwargs = {"rho0": 1.0, "Rs": 1, "r_core": 0.5}
        self.assert_integrals(Model, kwargs)
        kwargs_lens = {"alpha_Rs": 1.0, "Rs": 5.0, "r_core": 0.5}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_cnfw_ellipse(self):
        from lenstronomy.LensModel.Profiles.cnfw_ellipse import CNFW_ELLIPSE as Model

        kwargs = {"rho0": 1.0, "Rs": 1, "r_core": 0.5, "e1": 0, "e2": 0}
        # self.assert_integrals(Model, kwargs)
        kwargs_lens = {"alpha_Rs": 1.0, "Rs": 5.0, "r_core": 0.5, "e1": 0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs_lens)

    def test_cored_density(self):
        from lenstronomy.LensModel.Profiles.cored_density import CoredDensity as Model

        kwargs = {"sigma0": 0.1, "r_core": 6.0}
        self.assert_integrals(Model, kwargs)
        self.assert_lens_integrals(Model, kwargs)

    def test_cored_density_2(self):
        from lenstronomy.LensModel.Profiles.cored_density_2 import (
            CoredDensity2 as Model,
        )

        kwargs = {"sigma0": 0.1, "r_core": 6.0}
        self.assert_integrals(Model, kwargs)
        self.assert_lens_integrals(Model, kwargs)

    def test_cored_density_exp(self):
        from lenstronomy.LensModel.Profiles.cored_density_exp import (
            CoredDensityExp as Model,
        )

        kwargs = {"kappa_0": 0.1, "theta_c": 6.0}
        self.assert_integrals(Model, kwargs)
        self.assert_lens_integrals(Model, kwargs)

    def test_uldm(self):
        from lenstronomy.LensModel.Profiles.uldm import Uldm as Model

        kwargs = {"kappa_0": 0.1, "theta_c": 6.0, "slope": 7.8}
        self.assert_integrals(Model, kwargs)
        self.assert_lens_integrals(Model, kwargs)

    def test_splcore(self):
        from lenstronomy.LensModel.Profiles.splcore import SPLCORE as Model

        kwargs = {"rho0": 1.0, "gamma": 3, "r_core": 0.1}
        self.assert_integrals(Model, kwargs)
        kwargs = {"sigma0": 1.0, "gamma": 3, "r_core": 0.1}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"rho0": 1.0, "gamma": 2, "r_core": 0.1}
        self.assert_integrals(Model, kwargs)
        kwargs = {"sigma0": 1.0, "gamma": 2, "r_core": 0.1}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"rho0": 1.0, "gamma": 2.5, "r_core": 0.1}
        self.assert_integrals(Model, kwargs)
        kwargs = {"sigma0": 1.0, "gamma": 2.5, "r_core": 0.1}
        self.assert_lens_integrals(Model, kwargs)

    def test_nie(self):
        from lenstronomy.LensModel.Profiles.nie import NIE as Model

        kwargs = {"theta_E": 0.7, "s_scale": 0.3, "e1": 0.0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs)

    def test_chameleon(self):
        from lenstronomy.LensModel.Profiles.chameleon import Chameleon as Model

        kwargs = {"alpha_1": 2.0, "w_c": 0.1, "w_t": 2.0, "e1": 0.0, "e2": 0}
        self.assert_lens_integrals(Model, kwargs)

        from lenstronomy.LensModel.Profiles.chameleon import DoubleChameleon as Model

        kwargs = {
            "alpha_1": 2.0,
            "ratio": 2,
            "w_c1": 0.2,
            "w_t1": 1,
            "e11": 0,
            "e21": 0,
            "w_c2": 0.5,
            "w_t2": 2,
            "e12": 0,
            "e22": 0,
        }
        self.assert_lens_integrals(Model, kwargs)

        from lenstronomy.LensModel.Profiles.chameleon import TripleChameleon as Model

        kwargs = {
            "alpha_1": 2.0,
            "ratio12": 2,
            "ratio13": 0.2,
            "w_c1": 0.2,
            "w_t1": 1,
            "e11": 0,
            "e21": 0,
            "w_c2": 0.5,
            "w_t2": 2,
            "e12": 0,
            "e22": 0,
            "w_c3": 2,
            "w_t3": 5,
            "e13": 0,
            "e23": 0,
        }
        self.assert_lens_integrals(Model, kwargs)

    def test_gnfw(self):
        from lenstronomy.LensModel.Profiles.general_nfw import GNFW as Model

        kwargs = {"alpha_Rs": 0.7, "Rs": 0.3, "gamma_inner": 1.0, "gamma_outer": 3.2}
        self.assert_lens_integrals(Model, kwargs)
        kwargs = {"alpha_Rs": 0.7, "Rs": 0.3, "gamma_inner": 2.5, "gamma_outer": 3.0}
        self.assert_lens_integrals(Model, kwargs)
        kwargs = {"alpha_Rs": 0.7, "Rs": 0.3, "gamma_inner": 0.5, "gamma_outer": 2.6}
        self.assert_lens_integrals(Model, kwargs)

    def test_tnfwc(self):
        from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC as Model

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 0.1, "r_trunc": 200.0}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"alpha_Rs": 4.0, "Rs": 2.0, "r_core": 1.1, "r_trunc": 100.0}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"alpha_Rs": 4.0, "Rs": 0.5, "r_core": 0.1, "r_trunc": 3.0}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"alpha_Rs": 4.0, "Rs": 0.5, "r_core": 0.4, "r_trunc": 3.0}
        self.assert_lens_integrals(Model, kwargs)

        kwargs = {"alpha_Rs": 4.0, "Rs": 0.5, "r_core": 0.9, "r_trunc": 3.0}
        self.assert_lens_integrals(Model, kwargs)

    """
    def test_sis(self):
        kwargs = {'theta_E': 0.5}
        from astrofunc.LensingProfiles.sis import SIS as Model
        self.assert_integrals(Model, kwargs)

    def test_sersic(self):
        kwargs = {'n_sersic': .5, 'r_eff': 1.5, 'k_eff': 0.3}
        from astrofunc.LensingProfiles.sersic import Sersic as Model
        self.assert_integrals(Model, kwargs)

    """


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



# -*- coding: utf-8 -*-
"""This module defines ``class GaussianEllipseKappa`` to compute the lensing properties
of an elliptical Gaussian profile with ellipticity in the convergence using the formulae
from Shajib (2019)."""

__author__ = "ajshajib"

import numpy as np
from scipy.special import wofz
from scipy.integrate import quad
from copy import deepcopy
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["GaussianEllipseKappa"]


class GaussianEllipseKappa(LensProfileBase):
    """This class contains functions to evaluate the derivative and hessian matrix of
    the deflection potential for an elliptical Gaussian convergence.

    The formulae are from Shajib (2019).
    """

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "sigma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, use_scipy_wofz=True, min_ellipticity=1e-5):
        """Setup which method to use the Faddeeva function and the ellipticity limit for
        spherical approximation.

        :param use_scipy_wofz: If ``True``, use ``scipy.special.wofz``.
        :type use_scipy_wofz: ``bool``
        :param min_ellipticity: Minimum allowed ellipticity. For ``q > 1 - min_ellipticity``, values for spherical case will be returned.
        :type min_ellipticity: ``float``
        """
        if use_scipy_wofz:
            self.w_f = wofz
        else:
            self.w_f = self.w_f_approx

        self.min_ellipticity = min_ellipticity
        self.spherical = GaussianKappa()
        super(GaussianEllipseKappa, self).__init__()

    def function(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the potential function for elliptical Gaussian convergence.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Potential for elliptical Gaussian convergence
        :rtype: ``float``, or ``numpy.array`` with shape equal to ``x.shape``
        """
        phi_g, q = param_util.ellipticity2phi_q(e1, e2)

        if q > 1 - self.min_ellipticity:
            return self.spherical.function(x, y, amp, sigma, center_x, center_y)

        # adjusting amplitude to make the notation compatible with the
        # formulae given in Shajib (2019).
        amp_ = amp / (2 * np.pi * sigma**2)

        # converting ellipticity definition from q*x^2 + y^2/q to q^2*x^2 + y^2
        sigma_ = sigma * np.sqrt(q)  # * q

        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_g)
        sin_phi = np.sin(phi_g)

        x_ = cos_phi * x_shift + sin_phi * y_shift
        y_ = -sin_phi * x_shift + cos_phi * y_shift

        _b = 1.0 / 2.0 / sigma_**2
        _p = np.sqrt(_b * q**2 / (1.0 - q**2))

        if isinstance(x_, int) or isinstance(x_, float):
            return self._num_integral(x_, y_, amp_, sigma_, _p, q)
        else:
            f_ = []
            for i in range(len(x_)):
                f_.append(self._num_integral(x_[i], y_[i], amp_, sigma_, _p, q))
            return np.array(f_)

    def _num_integral(self, x_, y_, amp_, sigma_, _p, q):
        """

        :param x_:
        :param y_:
        :param _p:
        :param q:
        :return:
        """

        def pot_real_line_integrand(_x):
            sig_func_re, sig_func_im = self.sigma_function(_p * _x, 0, q)

            alpha_x_ = (
                amp_
                * sigma_
                * self.sgn(_x)
                * np.sqrt(2 * np.pi / (1.0 - q**2))
                * sig_func_re
            )

            return alpha_x_

        def pot_imag_line_integrand(_y):
            sig_func_re, sig_func_im = self.sigma_function(_p * x_, _p * _y, q)

            alpha_y_ = (
                -amp_
                * sigma_
                * self.sgn(x_ + 1j * _y)
                * np.sqrt(2 * np.pi / (1.0 - q**2))
                * sig_func_im
            )

            return alpha_y_

        pot_on_real_line = quad(pot_real_line_integrand, 0, x_)[0]
        pot_on_imag_parallel = quad(pot_imag_line_integrand, 0, y_)[0]
        return pot_on_real_line + pot_on_imag_parallel

    def derivatives(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the derivatives of function angles :math:`\\partial f/\\partial x`,
        :math:`\\partial f/\\partial y` at :math:`x,\\ y`.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Deflection angle :math:`\\partial f/\\partial x`, :math:`\\partial f/\\partial y`
         for elliptical Gaussian convergence.
        :rtype: tuple ``(float, float)`` or ``(numpy.array, numpy.array)`` with each ``numpy.array``'s shape equal
         to ``x.shape``.
        """
        phi_g, q = param_util.ellipticity2phi_q(e1, e2)

        if q > 1 - self.min_ellipticity:
            return self.spherical.derivatives(x, y, amp, sigma, center_x, center_y)

        # adjusting amplitude to make the notation compatible with the
        # formulae given in Shajib (2019).
        amp_ = amp / (2 * np.pi * sigma**2)

        # converting ellipticity definition from q*x^2 + y^2/q to q^2*x^2 + y^2
        sigma_ = sigma * np.sqrt(q)  # * q

        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_g)
        sin_phi = np.sin(phi_g)

        # rotated coordinates
        x_ = cos_phi * x_shift + sin_phi * y_shift
        y_ = -sin_phi * x_shift + cos_phi * y_shift

        _p = q / sigma_ / np.sqrt(2 * (1.0 - q**2))

        sig_func_re, sig_func_im = self.sigma_function(_p * x_, _p * y_, q)

        alpha_x_ = (
            amp_
            * sigma_
            * self.sgn(x_ + 1j * y_)
            * np.sqrt(2 * np.pi / (1.0 - q**2))
            * sig_func_re
        )
        alpha_y_ = (
            -amp_
            * sigma_
            * self.sgn(x_ + 1j * y_)
            * np.sqrt(2 * np.pi / (1.0 - q**2))
            * sig_func_im
        )

        # rotate back to the original frame
        f_x = alpha_x_ * cos_phi - alpha_y_ * sin_phi
        f_y = alpha_x_ * sin_phi + alpha_y_ * cos_phi

        return f_x, f_y

    def hessian(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute Hessian matrix of function :math:`\\partial^2f/\\partial x^2`,
        :math:`\\partial^2 f/\\partial y^2`, :math:`\\partial^2/\\partial x\\partial y`.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :return: Hessian :math:`A/(2 \\pi \\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)` for elliptical Gaussian
         convergence.
        :rtype: tuple ``(float, float, float)`` , or ``(numpy.array, numpy.array, numpy.array)`` with each ``numpy.array``'s shape equal to ``x.shape``.
        """
        phi_g, q = param_util.ellipticity2phi_q(e1, e2)

        if q > 1 - self.min_ellipticity:
            return self.spherical.hessian(x, y, amp, sigma, center_x, center_y)

        # adjusting amplitude to make the notation compatible with the
        # formulae given in Shajib (2019).
        amp_ = amp / (2 * np.pi * sigma**2)

        # converting ellipticity definition from q*x^2 + y^2/q to q^2*x^2 + y^2
        sigma_ = sigma * np.sqrt(q)  # * q

        x_shift = x - center_x
        y_shift = y - center_y
        cos_phi = np.cos(phi_g)
        sin_phi = np.sin(phi_g)

        # rotated coordinates
        x_ = cos_phi * x_shift + sin_phi * y_shift
        y_ = -sin_phi * x_shift + cos_phi * y_shift

        _p = q / sigma_ / np.sqrt(2 * (1.0 - q**2))
        sig_func_re, sig_func_im = self.sigma_function(_p * x_, _p * y_, q)

        kappa = amp_ * np.exp(-(q**2 * x_**2 + y_**2) / 2 / sigma_**2)

        shear = (
            -1
            / (1 - q * q)
            * (
                (1 + q**2) * kappa
                - 2 * q * amp_
                + np.sqrt(2 * np.pi)
                * q
                * q
                * amp_
                * (x_ - 1j * y_)
                / sigma_
                / np.sqrt(1 - q * q)
                * (sig_func_re - 1j * sig_func_im)
            )
        )

        # in rotated frame
        f_xx_ = kappa + shear.real
        f_yy_ = kappa - shear.real
        f_xy_ = shear.imag

        # rotate back to the original frame
        f_xx = (
            f_xx_ * cos_phi**2 + f_yy_ * sin_phi**2 - 2 * sin_phi * cos_phi * f_xy_
        )
        f_yy = (
            f_xx_ * sin_phi**2 + f_yy_ * cos_phi**2 + 2 * sin_phi * cos_phi * f_xy_
        )
        f_xy = (
            sin_phi * cos_phi * (f_xx_ - f_yy_) + (cos_phi**2 - sin_phi**2) * f_xy_
        )

        return f_xx, f_xy, f_xy, f_yy

    def density_2d(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0):
        """Compute the density of elliptical Gaussian :math:`A/(2 \\pi \\sigma^2)
        \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`.

        :param x: x coordinate.
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate.
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``float``
        :param sigma: Standard deviation of Gaussian.
        :type sigma: ``float``
        :param e1: Ellipticity parameter 1.
        :type e1: ``float``
        :param e2: Ellipticity parameter 2.
        :type e2: ``float``
        :param center_x: x coordinate of centroid.
        :type center_x: ``float``
        :param center_y: y coordianate of centroid.
        :type center_y: ``float``
        :return: Density :math:`\\kappa` for elliptical Gaussian convergence.
        :rtype: ``float``, or ``numpy.array`` with shape = ``x.shape``.
        """
        f_xx, f_xy, f_yx, f_yy = self.hessian(
            x, y, amp, sigma, e1, e2, center_x, center_y
        )
        return (f_xx + f_yy) / 2

    @staticmethod
    def sgn(z):
        """Compute the sign function :math:`\\mathrm{sgn}(z)` factor for deflection as
        sugggested by Bray (1984). For current implementation, returning 1 is
        sufficient.

        :param z: Complex variable :math:`z = x + \\mathrm{i}y`
        :type z: ``complex``
        :return: :math:`\\mathrm{sgn}(z)`
        :rtype: ``float``
        """
        return 1.0
        # np.sqrt(z*z)/z #np.sign(z.real*z.imag)
        # return np.sign(z.real)
        # if z.real != 0:
        #    return np.sign(z.real)
        # else:
        #    return np.sign(z.imag)
        # return np.where(z.real == 0, np.sign(z.real), np.sign(z.imag))

    def sigma_function(self, x, y, q):
        r"""Compute the function :math:`\varsigma (z; q)` from equation (4.12) of Shajib
        (2019).

        :param x: Real part of complex variable, :math:`x = \mathrm{Re}(z)`
        :type x: ``float`` or ``numpy.array``
        :param y: Imaginary part of complex variable, :math:`y = \mathrm{Im}(z)`
        :type y: ``float`` or ``numpy.array``
        :param q: Axis ratio
        :type q: ``float``
        :return: real and imaginary part of :math:`\varsigma(z; q)` function
        :rtype: tuple ``(type(x), type(x))``
        """
        y_sign = np.sign(y)
        y_ = deepcopy(y) * y_sign
        z = x + 1j * y_
        zq = q * x + 1j * y_ / q

        w = self.w_f(z)
        wq = self.w_f(zq)

        # exponential factor in the 2nd term of eqn. (4.15) of Shajib (2019)
        exp_factor = np.exp(-x * x * (1 - q * q) - y_ * y_ * (1 / q / q - 1))

        sigma_func_real = w.imag - exp_factor * wq.imag
        sigma_func_imag = (-w.real + exp_factor * wq.real) * y_sign

        return sigma_func_real, sigma_func_imag

    @staticmethod
    def w_f_approx(z):
        """Compute the Faddeeva function :math:`w_{\\mathrm F}(z)` using the
        approximation given in Zaghloul (2017).

        :param z: complex number
        :type z: ``complex`` or ``numpy.array(dtype=complex)``
        :return: :math:`w_\\mathrm{F}(z)`
        :rtype: ``complex``
        """
        sqrt_pi = 1 / np.sqrt(np.pi)
        i_sqrt_pi = 1j * sqrt_pi

        wz = np.empty_like(z)

        z_imag2 = z.imag**2
        abs_z2 = z.real**2 + z_imag2

        reg1 = abs_z2 >= 38000.0
        if np.any(reg1):
            wz[reg1] = i_sqrt_pi / z[reg1]

        reg2 = (256.0 <= abs_z2) & (abs_z2 < 38000.0)
        if np.any(reg2):
            t = z[reg2]
            wz[reg2] = i_sqrt_pi * t / (t * t - 0.5)

        reg3 = (62.0 <= abs_z2) & (abs_z2 < 256.0)
        if np.any(reg3):
            t = z[reg3]
            wz[reg3] = (i_sqrt_pi / t) * (1 + 0.5 / (t * t - 1.5))

        reg4 = (30.0 <= abs_z2) & (abs_z2 < 62.0) & (z_imag2 >= 1e-13)
        if np.any(reg4):
            t = z[reg4]
            tt = t * t
            wz[reg4] = (i_sqrt_pi * t) * (tt - 2.5) / (tt * (tt - 3.0) + 0.75)

        reg5 = (
            (62.0 > abs_z2) & np.logical_not(reg4) & (abs_z2 > 2.5) & (z_imag2 < 0.072)
        )
        if np.any(reg5):
            t = z[reg5]
            u = -t * t
            f1 = sqrt_pi
            f2 = 1
            s1 = [1.320522, 35.7668, 219.031, 1540.787, 3321.99, 36183.31]
            s2 = [1.841439, 61.57037, 364.2191, 2186.181, 9022.228, 24322.84, 32066.6]

            for s in s1:
                f1 = s - f1 * u
            for s in s2:
                f2 = s - f2 * u

            wz[reg5] = np.exp(u) + 1j * t * f1 / f2

        reg6 = (30.0 > abs_z2) & np.logical_not(reg5)
        if np.any(reg6):
            t3 = -1j * z[reg6]

            f1 = sqrt_pi
            f2 = 1
            s1 = [5.9126262, 30.180142, 93.15558, 181.92853, 214.38239, 122.60793]
            s2 = [
                10.479857,
                53.992907,
                170.35400,
                348.70392,
                457.33448,
                352.73063,
                122.60793,
            ]

            for s in s1:
                f1 = f1 * t3 + s
            for s in s2:
                f2 = f2 * t3 + s

            wz[reg6] = f1 / f2
        return wz



from lenstronomy.Cosmo import micro_lensing
from lenstronomy.Util import constants
import numpy.testing as npt


def test_einstein_radius():
    # from Wikipedia, a 60 M_jupiter mass object at 4000 pc with a source at 8000pc results in an Einstein radius of
    # about 0.00024 arc seconds
    mass = 60 * constants.M_jupiter / constants.M_sun
    d_l = 4000
    d_s = 8000
    theta_E = micro_lensing.einstein_radius(mass=mass, d_l=d_l, d_s=d_s)
    npt.assert_almost_equal(theta_E / 0.00024, 1, decimal=2)


def test_source_size():
    size_arc_seconds = micro_lensing.source_size(diameter=1, d_s=1000)
    npt.assert_almost_equal(size_arc_seconds, 9.304945278694935e-06, decimal=9)



from lenstronomy.Sampling.Likelihoods.prior_likelihood import PriorLikelihood
import numpy.testing as npt
import numpy as np
import pytest


class TestImageLikelihood(object):
    def setup_method(self):
        self.prior = PriorLikelihood(
            prior_lens=[[0, "gamma", 2, 0.1]],
            prior_source=None,
            prior_lens_light=None,
            prior_ps=None,
            prior_special=[["source_size", 1, 0.1]],
        )
        self.prior_lognormal = PriorLikelihood(
            prior_lens_lognormal=[[0, "gamma", np.log(2.0), 0.1]],
            prior_source_lognormal=None,
            prior_lens_light_lognormal=None,
            prior_ps_lognormal=None,
            prior_special_lognormal=[["source_size", 0.0, 0.1]],
        )

    def test_logL(self):
        kwargs_lens = [{"gamma": 2.0}]
        kwargs_cosmo = {"source_size": 1.0}
        logL = self.prior.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
            kwargs_special=kwargs_cosmo,
        )
        assert logL == 0

        kwargs_lens = [{"gamma": 2.1}]
        kwargs_cosmo = {"source_size": 1.1}
        logL = self.prior.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
            kwargs_special=kwargs_cosmo,
        )
        npt.assert_almost_equal(logL, -1, decimal=8)

    def test_logL_lognormal(self):
        kwargs_lens = [{"gamma": 2}]
        kwargs_cosmo = {"source_size": 1}
        logL = self.prior_lognormal.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
            kwargs_special=kwargs_cosmo,
        )
        assert logL == -3.0

        kwargs_lens = [{"gamma": 2.1}]
        kwargs_cosmo = {"source_size": 1.1}
        logL = self.prior_lognormal.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
            kwargs_special=kwargs_cosmo,
        )
        npt.assert_almost_equal(logL, -3.7732255247006443, decimal=8)

    def gauss(self, x, mean, simga):
        return np.exp(-(((x - mean) / (simga)) ** 2) / 2) / np.sqrt(2 * np.pi) / simga

    def test_kde_prior(self):
        x_array = np.linspace(1.0, 3.0, 200)
        sigma = 0.2
        mean = 2
        # pdf_array = self.gauss(x_array, mean=mean, simga=sigma)
        sample = np.random.normal(loc=mean, scale=sigma, size=50000)
        # approx = Approx(x_array, pdf_array)
        # sample = approx.draw(n=50000)
        prior = PriorLikelihood(prior_lens_kde=[[0, "gamma", sample]])

        kwargs_lens = [{"gamma": 2}]
        logL = prior.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
        )

        kwargs_lens = [{"gamma": 2.2}]
        logL_sigma = prior.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
        )
        delta_log = logL - logL_sigma
        npt.assert_almost_equal(delta_log, 0.5, decimal=1)

        kwargs_lens = [{"gamma": 2.4}]
        logL_sigma = prior.logL(
            kwargs_lens=kwargs_lens,
            kwargs_source=[],
            kwargs_lens_light=[],
            kwargs_ps=[],
        )
        delta_log = logL - logL_sigma
        npt.assert_almost_equal(delta_log, 2, decimal=1)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.Workflow.multi_band_manager import MultiBandUpdateManager
import pytest


class TestUpdateManager(object):
    def setup_method(self):
        kwargs_model = {
            "lens_model_list": ["SHEAR", "SHEAR"],
            "source_light_model_list": ["UNIFORM"],
            "index_lens_model_list": [[0], [1]],
        }
        kwargs_constraints = {}
        kwargs_likelihood = {}
        kwargs_params = {}
        lens_init = [
            {"e1": 0, "e2": 0, "ra_0": 0, "dec_0": 0},
            {"e1": 0, "e2": 0, "ra_0": 0, "dec_0": 0},
        ]
        lens_sigma = [{"e1": 0.1, "e2": 0.1}, {"e1": 0.1, "e2": 0.1}]
        lens_fixed = [{"ra_0": 0, "dec_0": 0}, {"ra_0": 0, "dec_0": 0}]
        lens_lower = [{"e1": -1, "e2": -1}, {"e1": -1, "e2": -1}]
        lens_upper = [{"e1": 1, "e2": 1}, {"e1": 1, "e2": 1}]
        kwargs_params["lens_model"] = [
            lens_init,
            lens_sigma,
            lens_fixed,
            lens_lower,
            lens_upper,
        ]
        kwargs_params["source_model"] = [[{}], [{}], [{}], [{}], [{}]]
        kwargs_params["special"] = [
            {"special1": 1},
            {"special1": 1},
            {"special1": 0.1},
            {"special1": 0},
            {"special1": 1},
        ]
        self.manager = MultiBandUpdateManager(
            kwargs_model,
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_params,
            num_bands=2,
        )

    def test_none_mamager(self):
        manager = MultiBandUpdateManager(
            kwargs_model={},
            kwargs_constraints={},
            kwargs_likelihood={},
            kwargs_params={},
            num_bands=0,
        )
        results = manager.best_fit()
        assert len(results["kwargs_lens"]) == 0

    def test_keep_frame_fixed(self):
        frame_list_fixed = [0]
        assert "e1" not in self.manager._lens_fixed[0]
        self.manager.keep_frame_fixed(frame_list_fixed)
        assert "e1" in self.manager._lens_fixed[0]

        self.manager.undo_frame_fixed(frame_list=[0])
        assert "e1" not in self.manager._lens_fixed[0]
        assert "ra_0" in self.manager._lens_fixed[0]

    def test_fix_not_computed(self):
        self.manager.fix_not_computed(free_bands=[False, True])
        print(self.manager._lens_fixed)
        assert "e1" in self.manager._lens_fixed[0]
        assert "ra_0" in self.manager._lens_fixed[0]
        assert "e1" not in self.manager._lens_fixed[1]


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.MultiPlane.multi_plane import MultiPlane
from lenstronomy.LensModel.Profiles.nfw import NFW
from lenstronomy.Util.util import make_grid
import unittest


class TestLensModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.lensModel = LensModel(["GAUSSIAN"])
        self.kwargs = [
            {
                "amp": 1.0,
                "sigma_x": 2.0,
                "sigma_y": 2.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]

    def test_init(self):
        lens_model_list = [
            "FLEXION",
            "SIS_TRUNCATED",
            "SERSIC",
            "SERSIC_ELLIPSE_KAPPA",
            "SERSIC_ELLIPSE_GAUSS_DEC",
            "NFW_ELLIPSE_GAUSS_DEC",
            "SERSIC_ELLIPSE_POTENTIAL",
            "CTNFW_GAUSS_DEC",
            "PJAFFE",
            "PJAFFE_ELLIPSE",
            "HERNQUIST_ELLIPSE",
            "INTERPOL",
            "INTERPOL_SCALED",
            "SHAPELETS_POLAR",
            "DIPOLE",
            "GAUSSIAN_ELLIPSE_KAPPA",
            "GAUSSIAN_ELLIPSE_POTENTIAL",
            "MULTI_GAUSSIAN_KAPPA",
            "MULTI_GAUSSIAN_KAPPA_ELLIPSE",
            "CHAMELEON",
            "DOUBLE_CHAMELEON",
        ]

        lensModel = LensModel(lens_model_list)
        assert len(lensModel.lens_model_list) == len(lens_model_list)

        lens_model_list = ["NFW"]
        lensModel = LensModel(lens_model_list)
        x, y = 0.2, 1
        kwargs = [{"alpha_Rs": 1, "Rs": 0.5, "center_x": 0, "center_y": 0}]
        value = lensModel.potential(x, y, kwargs)
        nfw_interp = NFW(interpol=True)
        value_interp_lookup = nfw_interp.function(x, y, **kwargs[0])
        npt.assert_almost_equal(value, value_interp_lookup, decimal=4)

    def test_kappa(self):
        lensModel = LensModel(lens_model_list=["CONVERGENCE"])
        kappa_ext = 0.5
        kwargs = [{"kappa": kappa_ext}]
        output = lensModel.kappa(x=1.0, y=1.0, kwargs=kwargs)
        assert output == kappa_ext

    def test_potential(self):
        output = self.lensModel.potential(x=1.0, y=1.0, kwargs=self.kwargs)
        npt.assert_almost_equal(output, 0.77880078307140488 / (8 * np.pi), decimal=8)
        # assert output == 0.77880078307140488/(8*np.pi)

    def test_alpha(self):
        output1, output2 = self.lensModel.alpha(x=1.0, y=1.0, kwargs=self.kwargs)
        npt.assert_almost_equal(output1, -0.19470019576785122 / (8 * np.pi), decimal=8)
        npt.assert_almost_equal(output2, -0.19470019576785122 / (8 * np.pi), decimal=8)
        # assert output1 == -0.19470019576785122/(8*np.pi)
        # assert output2 == -0.19470019576785122/(8*np.pi)

        output1_diff, output2_diff = self.lensModel.alpha(
            x=1.0, y=1.0, kwargs=self.kwargs, diff=0.00001
        )
        npt.assert_almost_equal(output1_diff, output1, decimal=5)
        npt.assert_almost_equal(output2_diff, output2, decimal=5)

    def test_gamma(self):
        lensModel = LensModel(lens_model_list=["SHEAR"])
        gamma1, gamm2 = 0.1, -0.1
        kwargs = [{"gamma1": gamma1, "gamma2": gamm2}]
        e1_out, e2_out = lensModel.gamma(x=1.0, y=1.0, kwargs=kwargs)
        assert e1_out == gamma1
        assert e2_out == gamm2

        output1, output2 = self.lensModel.gamma(x=1.0, y=1.0, kwargs=self.kwargs)
        assert output1 == 0
        assert output2 == 0.048675048941962805 / (8 * np.pi)

    def test_magnification(self):
        output = self.lensModel.magnification(x=1.0, y=1.0, kwargs=self.kwargs)
        assert output == 0.98848384784633392

    def test_flexion(self):
        lensModel = LensModel(lens_model_list=["FLEXION"])
        g1, g2, g3, g4 = 0.01, 0.02, 0.03, 0.04
        kwargs = [{"g1": g1, "g2": g2, "g3": g3, "g4": g4}]
        f_xxx, f_xxy, f_xyy, f_yyy = lensModel.flexion(x=1.0, y=1.0, kwargs=kwargs)
        npt.assert_almost_equal(f_xxx, g1, decimal=8)
        npt.assert_almost_equal(f_xxy, g2, decimal=8)
        npt.assert_almost_equal(f_xyy, g3, decimal=8)
        npt.assert_almost_equal(f_yyy, g4, decimal=8)

    def test_ray_shooting(self):
        delta_x, delta_y = self.lensModel.ray_shooting(x=1.0, y=1.0, kwargs=self.kwargs)
        npt.assert_almost_equal(
            delta_x, 1 + 0.19470019576785122 / (8 * np.pi), decimal=8
        )
        npt.assert_almost_equal(
            delta_y, 1 + 0.19470019576785122 / (8 * np.pi), decimal=8
        )
        # assert delta_x == 1 + 0.19470019576785122/(8*np.pi)
        # assert delta_y == 1 + 0.19470019576785122/(8*np.pi)

    def test_arrival_time(self):
        z_lens = 0.5
        z_source = 1.5
        x_image, y_image = 1.0, 0.0
        lensModel = LensModel(
            lens_model_list=["SIS"],
            multi_plane=True,
            lens_redshift_list=[z_lens],
            z_source=z_source,
        )
        kwargs = [{"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0}]
        arrival_time_mp = lensModel.arrival_time(x_image, y_image, kwargs)
        lensModel_sp = LensModel(
            lens_model_list=["SIS"], z_source=z_source, z_lens=z_lens
        )
        arrival_time_sp = lensModel_sp.arrival_time(x_image, y_image, kwargs)
        npt.assert_almost_equal(arrival_time_sp, arrival_time_mp, decimal=8)

    def test_fermat_potential(self):
        z_lens = 0.5
        z_source = 1.5
        x_image, y_image = 1.0, 0.0
        lensModel = LensModel(
            lens_model_list=["SIS"],
            multi_plane=True,
            lens_redshift_list=[z_lens],
            z_lens=z_lens,
            z_source=z_source,
        )
        kwargs = [{"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0}]
        fermat_pot = lensModel.fermat_potential(x_image, y_image, kwargs)
        arrival_time = lensModel.arrival_time(x_image, y_image, kwargs)
        arrival_time_from_fermat_pot = lensModel._lensCosmo.time_delay_units(fermat_pot)
        npt.assert_almost_equal(arrival_time_from_fermat_pot, arrival_time, decimal=8)

    def test_curl(self):
        z_lens_list = [0.2, 0.8]
        z_source = 1.5
        lensModel = LensModel(
            lens_model_list=["SIS", "SIS"],
            multi_plane=True,
            lens_redshift_list=z_lens_list,
            z_source=z_source,
        )
        kwargs = [
            {"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0},
            {"theta_E": 0.0, "center_x": 0.0, "center_y": 0.2},
        ]

        curl = lensModel.curl(x=1, y=1, kwargs=kwargs)
        assert curl == 0

        kwargs = [
            {"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0},
            {"theta_E": 1.0, "center_x": 0.0, "center_y": 0.2},
        ]

        curl = lensModel.curl(x=1, y=1, kwargs=kwargs)
        assert curl != 0

    def test_hessian_differentials(self):
        """Routine to test the private numerical differentials, both cross and square
        methods in the infinitesimal regime."""
        lens_model = LensModel(lens_model_list=["SIS"])
        kwargs = [{"theta_E": 1, "center_x": 0.01, "center_y": 0}]
        x, y = make_grid(numPix=10, deltapix=0.2)
        diff = 0.0000001
        f_xx_sq, f_xy_sq, f_yx_sq, f_yy_sq = lens_model.hessian(
            x, y, kwargs, diff=diff, diff_method="square"
        )
        f_xx_cr, f_xy_cr, f_yx_cr, f_yy_cr = lens_model.hessian(
            x, y, kwargs, diff=diff, diff_method="cross"
        )
        f_xx, f_xy, f_yx, f_yy = lens_model.hessian(x, y, kwargs, diff=None)
        npt.assert_almost_equal(f_xx_cr, f_xx, decimal=5)
        npt.assert_almost_equal(f_xy_cr, f_xy, decimal=5)
        npt.assert_almost_equal(f_yx_cr, f_yx, decimal=5)
        npt.assert_almost_equal(f_yy_cr, f_yy, decimal=5)

        npt.assert_almost_equal(f_xx_sq, f_xx, decimal=5)
        npt.assert_almost_equal(f_xy_sq, f_xy, decimal=5)
        npt.assert_almost_equal(f_yx_sq, f_yx, decimal=5)
        npt.assert_almost_equal(f_yy_sq, f_yy, decimal=5)

    def test_hessian_z1z2(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        kwargs_lens = [{"theta_E": 1}]
        redshift_list = [0.5]
        lensModel = LensModel(
            lens_model_list=lens_model_list,
            multi_plane=True,
            lens_redshift_list=redshift_list,
            z_source=z_source,
        )
        z1, z2 = 0.5, 1.5
        theta_x, theta_y = np.linspace(start=-1, stop=1, num=10), np.linspace(
            start=-1, stop=1, num=10
        )

        f_xx, f_xy, f_yx, f_yy = lensModel.hessian_z1z2(
            z1, z2, theta_x, theta_y, kwargs_lens
        )
        # Use the method in multi_plane.hessian_z1z2 as a comparison
        multi_plane = MultiPlane(
            z_source=1.5,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            z_interp_stop=3,
            cosmo_interp=False,
        )
        (
            f_xx_expected,
            f_xy_expected,
            f_yx_expected,
            f_yy_expected,
        ) = multi_plane.hessian_z1z2(
            z1=z1, z2=z2, theta_x=theta_x, theta_y=theta_y, kwargs_lens=kwargs_lens
        )
        npt.assert_almost_equal(f_xx, f_xx_expected, decimal=5)
        npt.assert_almost_equal(f_xy, f_xy_expected, decimal=5)
        npt.assert_almost_equal(f_yx, f_yx_expected, decimal=5)
        npt.assert_almost_equal(f_yy, f_yy_expected, decimal=5)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            kwargs = [{"alpha_Rs": 1, "Rs": 0.5, "center_x": 0, "center_y": 0}]
            lensModel = LensModel(
                ["NFW"], multi_plane=True, lens_redshift_list=[1], z_source=2
            )
            f_x, f_y = lensModel.alpha(1, 1, kwargs, diff=0.0001)
        with self.assertRaises(ValueError):
            lensModel = LensModel(["NFW"], multi_plane=True, lens_redshift_list=[1])
        with self.assertRaises(ValueError):
            kwargs = [{"alpha_Rs": 1, "Rs": 0.5, "center_x": 0, "center_y": 0}]
            lensModel = LensModel(["NFW"], multi_plane=False)
            t_arrival = lensModel.arrival_time(1, 1, kwargs)
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            x_image, y_image = 1.0, 0.0
            lensModel = LensModel(
                lens_model_list=["SIS"],
                multi_plane=True,
                lens_redshift_list=[z_lens],
                z_source=z_source,
            )
            kwargs = [{"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0}]
            fermat_pot = lensModel.fermat_potential(x_image, y_image, kwargs)
        with self.assertRaises(ValueError):
            lens_model = LensModel(lens_model_list=["SIS"])
            kwargs = [{"theta_E": 1.0, "center_x": 0.0, "center_y": 0.0}]
            lens_model.hessian(0, 0, kwargs, diff=0.001, diff_method="bad")
        with self.assertRaises(ValueError):
            lens_model = LensModel(lens_model_list=["LOS", "LOS_MINIMAL"])
        with self.assertRaises(ValueError):
            lens_model = LensModel(
                lens_model_list=["LOS", "EPL", "NFW"], multi_plane=True, z_source=1.0
            )
        with self.assertRaises(ValueError):
            lens_model = LensModel(
                lens_model_list=["LOS_MINIMAL", "SIS", "GAUSSIAN"],
                multi_plane=True,
                z_source=1.0,
            )

    def test_hessian_z1z2_raise(self):
        lensModel = LensModel(
            lens_model_list=["SIS"],
            multi_plane=True,
            lens_redshift_list=[1],
            z_source=2,
        )
        kwargs = [{"theta_E": 1, "center_x": 0, "center_y": 0}]

        # Test when the model is not in multi-plane mode
        lensModel_non_multi = LensModel(
            lens_model_list=["SIS"],
            multi_plane=False,
            lens_redshift_list=[1],
            z_source=2,
        )
        with self.assertRaises(ValueError):
            lensModel_non_multi.hessian_z1z2(0.5, 1.5, 1, 1, kwargs)

        # Test when z1 >= z2
        with self.assertRaises(ValueError):
            lensModel.hessian_z1z2(1.5, 1.5, 1, 1, kwargs)
        with self.assertRaises(ValueError):
            lensModel.hessian_z1z2(2.0, 1.5, 1, 1, kwargs)


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



__author__ = "gipagano"


import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.Util import util
from lenstronomy.LensModel.Profiles.const_mag import ConstMag


class TestCONST_MAG(object):
    """Tests the CONST_MAG profile for different rotations."""

    def setup_method(self):
        self.const_mag = ConstMag()

    def test_function(self):
        y = np.array([1.0, 2])
        x = np.array([0.0, 0.0])

        mu_r = 1.0
        mu_t = 10.0

        # positive parity

        parity = 1

        ############
        # rotation 1
        ############

        phi_G = np.pi

        values = self.const_mag.function(x, y, mu_r, mu_t, parity, phi_G)
        delta_pot = values[1] - values[0]

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f_ = self.const_mag.function(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back

        delta_pot_rot = f_[1] - f_[0]

        # compare
        npt.assert_almost_equal(delta_pot, delta_pot_rot, decimal=4)

        ############
        # rotation 2
        ############

        phi_G = np.pi / 3.0

        values = self.const_mag.function(x, y, mu_r, mu_t, parity, phi_G)
        delta_pot = values[1] - values[0]

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f_ = self.const_mag.function(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back

        delta_pot_rot = f_[1] - f_[0]

        # compare
        npt.assert_almost_equal(delta_pot, delta_pot_rot, decimal=4)

        # ===========================================================

        # negative parity

        parity = -1

        ############
        # rotation 1
        ############

        phi_G = np.pi

        values = self.const_mag.function(x, y, mu_r, mu_t, parity, phi_G)
        delta_pot = values[1] - values[0]

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f_ = self.const_mag.function(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back

        delta_pot_rot = f_[1] - f_[0]

        # compare
        npt.assert_almost_equal(delta_pot, delta_pot_rot, decimal=4)

        ############
        # rotation 2
        ############

        phi_G = np.pi / 3.0

        values = self.const_mag.function(x, y, mu_r, mu_t, parity, phi_G)
        delta_pot = values[1] - values[0]

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f_ = self.const_mag.function(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back

        delta_pot_rot = f_[1] - f_[0]

        # compare
        npt.assert_almost_equal(delta_pot, delta_pot_rot, decimal=4)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])

        mu_r = 1.0
        mu_t = 10.0

        # positive parity

        parity = 1

        ############
        # rotation 1
        ############

        phi_G = np.pi

        f_x, f_y = self.const_mag.derivatives(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__x, f__y = self.const_mag.derivatives(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back
        f_x_rot, f_y_rot = util.rotate(f__x, f__y, -phi_G)

        # compare
        npt.assert_almost_equal(f_x, f_x_rot, decimal=4)
        npt.assert_almost_equal(f_y, f_y_rot, decimal=4)

        ############
        # rotation 2
        ############

        phi_G = np.pi / 3.0

        f_x, f_y = self.const_mag.derivatives(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__x, f__y = self.const_mag.derivatives(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back
        f_x_rot, f_y_rot = util.rotate(f__x, f__y, -phi_G)

        # compare
        npt.assert_almost_equal(f_x, f_x_rot, decimal=4)
        npt.assert_almost_equal(f_y, f_y_rot, decimal=4)

        # ===========================================================

        # negative parity

        parity = -1

        ############
        # rotation 1
        ############

        phi_G = np.pi

        f_x, f_y = self.const_mag.derivatives(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__x, f__y = self.const_mag.derivatives(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back
        f_x_rot, f_y_rot = util.rotate(f__x, f__y, -phi_G)

        # compare
        npt.assert_almost_equal(f_x, f_x_rot, decimal=4)
        npt.assert_almost_equal(f_y, f_y_rot, decimal=4)

        ############
        # rotation 2
        ############

        phi_G = np.pi / 3.0

        f_x, f_y = self.const_mag.derivatives(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__x, f__y = self.const_mag.derivatives(x__, y__, mu_r, mu_t, parity, 0.0)

        # rotate back
        f_x_rot, f_y_rot = util.rotate(f__x, f__y, -phi_G)

        # compare
        npt.assert_almost_equal(f_x, f_x_rot, decimal=4)
        npt.assert_almost_equal(f_y, f_y_rot, decimal=4)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])

        mu_r = 1.0
        mu_t = 10.0

        # positive parity

        parity = 1

        ############
        # rotation 1
        ############

        phi_G = np.pi

        f_xx, f_xy, f_yx, f_yy = self.const_mag.hessian(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__xx, f__xy, f__yx, f__yy = self.const_mag.hessian(
            x__, y__, mu_r, mu_t, parity, 0.0
        )

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_G) * gamma1__ - np.sin(2 * phi_G) * gamma2__
        gamma2 = +np.sin(2 * phi_G) * gamma1__ + np.cos(2 * phi_G) * gamma2__
        f_xx_rot = kappa + gamma1
        f_yy_rot = kappa - gamma1
        f_xy_rot = gamma2

        # compare
        npt.assert_almost_equal(f_xx, f_xx_rot, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_rot, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_rot, decimal=4)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

        ############
        # rotation 2
        ############

        phi_G = np.pi / 3.0

        f_xx, f_xy, f_yx, f_yy = self.const_mag.hessian(x, y, mu_r, mu_t, parity, phi_G)

        # rotate
        x__, y__ = util.rotate(x, y, phi_G)

        # evaluate
        f__xx, f__xy, f__yx, f__yy = self.const_mag.hessian(
            x__, y__, mu_r, mu_t, parity, 0.0
        )

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_G) * gamma1__ - np.sin(2 * phi_G) * gamma2__
        gamma2 = +np.sin(2 * phi_G) * gamma1__ + np.cos(2 * phi_G) * gamma2__
        f_xx_rot = kappa + gamma1
        f_yy_rot = kappa - gamma1
        f_xy_rot = gamma2

        # compare
        npt.assert_almost_equal(f_xx, f_xx_rot, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_rot, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_rot, decimal=4)
        npt.assert_almost_equal(f_yx, f_xy_rot, decimal=4)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
import numpy as np
from lenstronomy.Util import param_util
from lenstronomy.Util import util

__all__ = [
    "CSE",
    "CSEMajorAxis",
    "CSEMajorAxisSet",
    "CSEProductAvg",
    "CSEProductAvgSet",
]


class CSE(LensProfileBase):
    """
    Cored steep ellipsoid (CSE)
    :param axis: 'major' or 'product_avg' ; whether to evaluate corresponding to r= major axis or r= sqrt(ab)
    source:
    Keeton and Kochanek (1998)
    Oguri 2021: https://arxiv.org/pdf/2106.11464.pdf

    .. math::
        \\kappa(u;s) = \\frac{A}{2(s^2 + \\xi^2)^{3/2}}

    with

    .. math::
        \\xi(x, y) = \\sqrt{x^2 + \\frac{y^2}{q^2}}

    """

    param_names = ["A", "s", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "A": -1000,
        "s": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "A": 1000,
        "s": 10000,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": -100,
        "center_y": -100,
    }

    def __init__(self, axis="product_avg"):
        if axis == "major":
            self.major_axis_model = CSEMajorAxis()
        elif axis == "product_avg":
            self.major_axis_model = CSEProductAvg()
        else:
            raise ValueError(
                "axis must be set to 'major' or 'product_avg'. Input is %s ." % axis
            )
        super(CSE, self).__init__()

    def function(self, x, y, a, s, e1, e2, center_x, center_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: center of profile
        :param center_y: center of profile
        :return: lensing potential
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)

        # potential calculation
        f_ = self.major_axis_model.function(x__, y__, a, s, q)
        return f_

    def derivatives(self, x, y, a, s, e1, e2, center_x, center_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: center of profile
        :param center_y: center of profile
        :return: deflection in x- and y-direction
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)

        f__x, f__y = self.major_axis_model.derivatives(x__, y__, a, s, q)

        # rotate deflections back
        f_x, f_y = util.rotate(f__x, f__y, -phi_q)
        return f_x, f_y

    def hessian(self, x, y, a, s, e1, e2, center_x, center_y):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param e1: eccentricity
        :param e2: eccentricity
        :param center_x: center of profile
        :param center_y: center of profile
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        phi_q, q = param_util.ellipticity2phi_q(e1, e2)
        # shift
        x_ = x - center_x
        y_ = y - center_y
        # rotate
        x__, y__ = util.rotate(x_, y_, phi_q)

        f__xx, f__xy, __, f__yy = self.major_axis_model.hessian(x__, y__, a, s, q)

        # rotate back
        kappa = 1.0 / 2 * (f__xx + f__yy)
        gamma1__ = 1.0 / 2 * (f__xx - f__yy)
        gamma2__ = f__xy
        gamma1 = np.cos(2 * phi_q) * gamma1__ - np.sin(2 * phi_q) * gamma2__
        gamma2 = +np.sin(2 * phi_q) * gamma1__ + np.cos(2 * phi_q) * gamma2__
        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2

        return f_xx, f_xy, f_xy, f_yy


class CSEMajorAxis(LensProfileBase):
    """
    Cored steep ellipsoid (CSE) along the major axis
    source:
    Keeton and Kochanek (1998)
    Oguri 2021: https://arxiv.org/pdf/2106.11464.pdf

    .. math::
        \\kappa(u;s) = \\frac{A}{2(s^2 + \\xi^2)^{3/2}}

    with

    .. math::
        \\xi(x, y) = \\sqrt{x^2 + \\frac{y^2}{q^2}}

    """

    param_names = ["A", "s", "q", "center_x", "center_y"]
    lower_limit_default = {
        "A": -1000,
        "s": 0,
        "q": 0.001,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "A": 1000,
        "s": 10000,
        "q": 0.99999,
        "e2": 0.5,
        "center_x": -100,
        "center_y": -100,
    }

    def function(self, x, y, a, s, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: lensing potential
        """

        # potential calculation
        psi = np.sqrt(q**2 * (s**2 + x**2) + y**2)
        Phi = (psi + s) ** 2 + (1 - q**2) * x**2
        phi = q / (2 * s) * np.log(Phi) - q / s * np.log((1 + q) * s)
        return a * phi

    def derivatives(self, x, y, a, s, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: deflection in x- and y-direction
        """

        psi = np.sqrt(q**2 * (s**2 + x**2) + y**2)
        Phi = (psi + s) ** 2 + (1 - q**2) * x**2
        f_x = q * x * (psi + q**2 * s) / (s * psi * Phi)
        f_y = q * y * (psi + s) / (s * psi * Phi)

        return a * f_x, a * f_y

    def hessian(self, x, y, a, s, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """

        # equations 21-23 in Oguri 2021
        psi = np.sqrt(q**2 * (s**2 + x**2) + y**2)
        Phi = (psi + s) ** 2 + (1 - q**2) * x**2
        f_xx = (
            q
            / (s * Phi)
            * (
                1
                + q**2 * s * (q**2 * s**2 + y**2) / psi**3
                - 2 * x**2 * (psi + q**2 * s) ** 2 / (psi**2 * Phi)
            )
        )
        f_yy = (
            q
            / (s * Phi)
            * (
                1
                + q**2 * s * (s**2 + x**2) / psi**3
                - 2 * y**2 * (psi + s) ** 2 / (psi**2 * Phi)
            )
        )
        f_xy = (
            -q
            * x
            * y
            / (s * Phi)
            * (
                q**2 * s / psi**3
                + 2 * (psi + q**2 * s) * (psi + s) / (psi**2 * Phi)
            )
        )

        return a * f_xx, a * f_xy, a * f_xy, a * f_yy


class CSEMajorAxisSet(LensProfileBase):
    """A set of CSE profiles along a joint center and axis."""

    def __init__(self):
        self.major_axis_model = CSEMajorAxis()
        super(CSEMajorAxisSet, self).__init__()

    def function(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: lensing potential
        """
        f_ = np.zeros_like(x)
        for a, s in zip(a_list, s_list):
            f_ += self.major_axis_model.function(x, y, a, s, q)
        return f_

    def derivatives(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: deflection in x- and y-direction
        """
        f_x, f_y = np.zeros_like(x), np.zeros_like(y)
        for a, s in zip(a_list, s_list):
            f_x_, f_y_ = self.major_axis_model.derivatives(x, y, a, s, q)
            f_x += f_x_
            f_y += f_y_
        return f_x, f_y

    def hessian(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        f_xx, f_xy, f_yy = np.zeros_like(x), np.zeros_like(x), np.zeros_like(x)
        for a, s in zip(a_list, s_list):
            f_xx_, f_xy_, _, f_yy_ = self.major_axis_model.hessian(x, y, a, s, q)
            f_xx += f_xx_
            f_xy += f_xy_
            f_yy += f_yy_
        return f_xx, f_xy, f_xy, f_yy


class CSEProductAvg(LensProfileBase):
    """Cored steep ellipsoid (CSE) evaluated at the product-averaged radius sqrt(ab),
    such that mass is not changed when increasing ellipticity.

    Same as CSEMajorAxis but evaluated at r=sqrt(q)*r_original

    Keeton and Kochanek (1998)
    Oguri 2021: https://arxiv.org/pdf/2106.11464.pdf

    .. math::
        \\kappa(u;s) = \\frac{A}{2(s^2 + \\xi^2)^{3/2}}

    with

    .. math::
        \\xi(x, y) = \\sqrt{qx^2 + \\frac{y^2}{q}}
    """

    param_names = ["A", "s", "q", "center_x", "center_y"]
    lower_limit_default = {
        "A": -1000,
        "s": 0,
        "q": 0.001,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "A": 1000,
        "s": 10000,
        "q": 0.99999,
        "e2": 0.5,
        "center_x": -100,
        "center_y": -100,
    }

    def __init__(self):
        super(CSEProductAvg, self).__init__()
        self.MA_class = CSEMajorAxis()

    @staticmethod
    def _convert2prodavg(x, y, a, s, q):
        """Converts coordinates and re-normalizes major-axis parameterization to instead
        be wrt.

        product-averaged
        """
        a = a / q
        x = x * np.sqrt(q)
        y = y * np.sqrt(q)
        return x, y, a, s, q

    def function(self, x, y, a, s, q):
        """
        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: lensing potential
        """
        x, y, a, s, q = self._convert2prodavg(x, y, a, s, q)
        return self.MA_class.function(x, y, a, s, q)

    def derivatives(self, x, y, a, s, q):
        """
        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: deflection in x- and y-direction
        """
        x, y, a, s, q = self._convert2prodavg(x, y, a, s, q)
        af_x, af_y = self.MA_class.derivatives(x, y, a, s, q)
        # extra sqrt(q) factor from taking derivative of transformed coordinate
        return np.sqrt(q) * af_x, np.sqrt(q) * af_y

    def hessian(self, x, y, a, s, q):
        """
        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a: lensing strength
        :param s: core radius
        :param q: axis ratio
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        x, y, a, s, q = self._convert2prodavg(x, y, a, s, q)
        af_xx, af_xy, af_xy, af_yy = self.MA_class.hessian(x, y, a, s, q)
        # two sqrt(q) factors from taking derivatives of transformed coordinate
        return q * af_xx, q * af_xy, q * af_xy, q * af_yy


class CSEProductAvgSet(LensProfileBase):
    """A set of CSE profiles along a joint center and axis."""

    def __init__(self):
        self.major_axis_model = CSEProductAvg()
        super(CSEProductAvgSet, self).__init__()

    def function(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: lensing potential
        """
        f_ = np.zeros_like(x)
        for a, s in zip(a_list, s_list):
            f_ += self.major_axis_model.function(x, y, a, s, q)
        return f_

    def derivatives(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: deflection in x- and y-direction
        """
        f_x, f_y = np.zeros_like(x), np.zeros_like(y)
        for a, s in zip(a_list, s_list):
            f_x_, f_y_ = self.major_axis_model.derivatives(x, y, a, s, q)
            f_x += f_x_
            f_y += f_y_
        return f_x, f_y

    def hessian(self, x, y, a_list, s_list, q):
        """

        :param x: coordinate in image plane (angle)
        :param y: coordinate in image plane (angle)
        :param a_list: list of lensing strength
        :param s_list: list of core radius
        :param q: axis ratio
        :return: hessian elements f_xx, f_xy, f_yx, f_yy
        """
        f_xx, f_xy, f_yy = np.zeros_like(x), np.zeros_like(x), np.zeros_like(x)
        for a, s in zip(a_list, s_list):
            f_xx_, f_xy_, _, f_yy_ = self.major_axis_model.hessian(x, y, a, s, q)
            f_xx += f_xx_
            f_xy += f_xy_
            f_yy += f_yy_
        return f_xx, f_xy, f_xy, f_yy



import numpy as np
import numpy.testing as npt
import scipy.signal
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.LensModel.lens_model import LensModel
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.Util import kernel_util
import lenstronomy.Util.util as util

from lenstronomy.ImSim.image_linear_solve import ImageLinearFit

"""
Test the linear solver for natwt (natural weighting) interferometric data.
Test the _image_linear_solve function of ImageLinearFit class.
The idea is to define data, psf, source, lens, lens light classes respectively, and run the linear solving
inside and outside of the _image_linear_solve function. Verify the 1st and 4th output of _image_linear_solve.
The test should be independent of the specific definitions of the light and lens profiles.
"""


def test_image_linear_solve_with_primary_beam_and_interferometry_psf():
    background_rms = 0.05
    exp_time = np.inf
    numPix = 80
    deltaPix = 0.05
    psf_type = "PIXEL"
    kernel_size = 161

    # simulate a primary beam (pb)
    primary_beam = np.zeros((numPix, numPix))
    for i in range(numPix):
        for j in range(numPix):
            primary_beam[i, j] = np.exp(-1e-4 * ((i - 78) ** 2 + (j - 56) ** 2))
    primary_beam /= np.max(primary_beam)

    # simulate a spherical sinc function as psf, which contains negative pixels
    psf_test = np.zeros((221, 221))
    for i in range(221):
        for j in range(221):
            if i > j:
                psf_test[i, j] = psf_test[j, i]
            r = np.sqrt((i - 110) ** 2 + (j - 110) ** 2)
            if r == 0:
                psf_test[i, j] = 1
            else:
                psf_test[i, j] = np.sin(r * 0.5) / (r * 0.5)

    # note that the simulated noise here is not the interferometric noise. we just use it to test the numerics
    test_noise = scipy.signal.fftconvolve(
        np.random.normal(0, 1, (numPix, numPix)), psf_test, mode="same"
    )

    kwargs_data = sim_util.data_configure_simple(
        numPix, deltaPix, exp_time, background_rms
    )
    kwargs_data["ra_at_xy_0"] = -(40) * deltaPix
    kwargs_data["dec_at_xy_0"] = -(40) * deltaPix
    kwargs_data["antenna_primary_beam"] = primary_beam
    kwargs_data[
        "likelihood_method"
    ] = "interferometry_natwt"  # testing just for interferometry natwt method
    data_class = ImageData(**kwargs_data)

    kernel_cut = kernel_util.cut_psf(psf_test, kernel_size, normalisation=False)
    kwargs_psf = {
        "psf_type": psf_type,
        "pixel_size": deltaPix,
        "kernel_point_source": kernel_cut,
        "kernel_point_source_normalisation": False,
    }
    psf_class = PSF(**kwargs_psf)

    # define lens model and source model
    kwargs_shear = {"gamma1": 0.01, "gamma2": 0.01}
    kwargs_spemd = {
        "theta_E": 1.0,
        "gamma": 1.8,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.04,
    }
    lens_model_list = ["SPEP", "SHEAR"]
    kwargs_lens = [kwargs_spemd, kwargs_shear]
    lens_model_class = LensModel(lens_model_list=lens_model_list)

    kwargs_sersic = {
        "amp": 25.0,
        "R_sersic": 0.3,
        "n_sersic": 2,
        "center_x": 0,
        "center_y": 0,
    }
    lens_light_model_list = ["SERSIC"]
    kwargs_lens_light = [kwargs_sersic]
    lens_light_model_class = LightModel(light_model_list=lens_light_model_list)

    kwargs_sersic_ellipse = {
        "amp": 10.0,
        "R_sersic": 0.6,
        "n_sersic": 7,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.05,
        "e2": 0.02,
    }
    source_model_list = ["SERSIC_ELLIPSE"]
    kwargs_source = [kwargs_sersic_ellipse]
    source_model_class = LightModel(light_model_list=source_model_list)

    kwargs_numerics = {"supersampling_factor": 1, "supersampling_convolution": False}

    imageModel = ImageModel(
        data_class,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    image_sim = imageModel.image(kwargs_lens, kwargs_source, kwargs_lens_light)

    # normalize the noise to make it small compared to the model image
    test_noise *= 1e-2 * (np.max(image_sim) / np.std(test_noise))
    sim_data = image_sim + test_noise
    data_class.update_data(sim_data)

    # define the ImageLinearFit class using the materials defined above, run the _image_linear_solve function
    imageLinearFit = ImageLinearFit(
        data_class,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    model, _, _, amps = imageLinearFit._image_linear_solve(
        kwargs_lens, kwargs_source, kwargs_lens_light
    )

    # execute the same linear solving outside of the _image_linear_solve function
    A = imageLinearFit._linear_response_matrix(
        kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps=None, unconvolved=True
    )
    A0 = util.array2image(A[0])
    A1 = util.array2image(A[1])
    A0c = scipy.signal.fftconvolve(A0, psf_test, mode="same")
    A1c = scipy.signal.fftconvolve(A1, psf_test, mode="same")
    M = np.zeros((2, 2))
    b = np.zeros((2))
    M[0, 0] = np.sum(A0c * A0)
    M[0, 1] = np.sum(A0c * A1)
    M[1, 0] = np.sum(A1c * A0)
    M[1, 1] = np.sum(A1c * A1)
    b[0] = np.sum(A0 * sim_data)
    b[1] = np.sum(A1 * sim_data)

    amps0 = np.linalg.lstsq(M, b)[0]
    clean_model = amps0[0] * A0 + amps0[1] * A1
    dirty_model = amps0[0] * A0c + amps0[1] * A1c

    npt.assert_almost_equal([clean_model, dirty_model], model, decimal=8)
    npt.assert_almost_equal(amps0, amps, decimal=8)



import pytest
import numpy as np
import numpy.testing as npt
import unittest

from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
import lenstronomy.Util.param_util as param_util


class TestPointSource(object):
    def setup_method(self):
        lensModel = LensModel(lens_model_list=["SPEP"])
        solver = LensEquationSolver(lensModel=lensModel)
        e1, e2 = param_util.phi_q2_ellipticity(0, 0.7)
        self.kwargs_lens = [
            {
                "theta_E": 1.0,
                "center_x": 0,
                "center_y": 0,
                "e1": e1,
                "e2": e2,
                "gamma": 2,
            }
        ]
        self.sourcePos_x, self.sourcePos_y = 0.01, -0.01
        self.x_pos, self.y_pos = solver.image_position_from_source(
            sourcePos_x=self.sourcePos_x,
            sourcePos_y=self.sourcePos_y,
            kwargs_lens=self.kwargs_lens,
        )
        self.PointSource = PointSource(
            point_source_type_list=["LENSED_POSITION", "UNLENSED", "SOURCE_POSITION"],
            lensModel=lensModel,
            fixed_magnification_list=[False] * 3,
            additional_images_list=[False] * 4,
            flux_from_point_source_list=[True, True, True],
            index_lens_model_list=[[0]],
            point_source_frame_list=[[0] * len(self.x_pos), [0], [0]],
        )
        self.kwargs_ps = [
            {
                "ra_image": self.x_pos,
                "dec_image": self.y_pos,
                "point_amp": np.ones_like(self.x_pos) * 2,
            },
            {"ra_image": [1.0], "dec_image": [1.0], "point_amp": [10]},
            {
                "ra_source": self.sourcePos_x,
                "dec_source": self.sourcePos_y,
                "point_amp": np.ones_like(self.x_pos),
            },
            {},
        ]

    def test_image_position(self):
        x_image_list, y_image_list = self.PointSource.image_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(x_image_list[0][0], self.x_pos[0], decimal=8)
        npt.assert_almost_equal(x_image_list[1], 1, decimal=8)
        npt.assert_almost_equal(x_image_list[2][0], self.x_pos[0], decimal=8)

        x_image_list, y_image_list = self.PointSource.image_position(
            kwargs_ps=self.kwargs_ps,
            kwargs_lens=self.kwargs_lens,
            original_position=True,
            additional_images=True,
        )
        npt.assert_almost_equal(x_image_list[0][0], self.x_pos[0], decimal=8)
        npt.assert_almost_equal(x_image_list[1], 1, decimal=8)
        npt.assert_almost_equal(x_image_list[2][0], self.x_pos[0], decimal=8)

    def test_source_position(self):
        x_source_list, y_source_list = self.PointSource.source_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(x_source_list[0], self.sourcePos_x, decimal=8)
        npt.assert_almost_equal(x_source_list[1], 1, decimal=8)
        npt.assert_almost_equal(x_source_list[2], self.sourcePos_x, decimal=8)

    def test_num_basis(self):
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert num_basis == 9

    def test_linear_response_set(self):
        ra_pos, dec_pos, amp, n = self.PointSource.linear_response_set(
            self.kwargs_ps, kwargs_lens=self.kwargs_lens, with_amp=False
        )
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert amp[0][0] == 1
        assert n == num_basis
        assert ra_pos[0][0] == self.x_pos[0]

        ra_pos, dec_pos, amp, n = self.PointSource.linear_response_set(
            self.kwargs_ps, kwargs_lens=self.kwargs_lens, with_amp=True
        )
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert amp[0][0] != 1
        assert n == num_basis
        assert ra_pos[0][0] == self.x_pos[0]

    def test_linear_param_from_kwargs(self):
        param = self.PointSource.linear_param_from_kwargs(self.kwargs_ps)
        assert param[0] == self.kwargs_ps[0]["point_amp"][0]
        assert param[1] == self.kwargs_ps[0]["point_amp"][1]

    def test_point_source_list(self):
        ra_list, dec_list, amp_list = self.PointSource.point_source_list(
            self.kwargs_ps, self.kwargs_lens
        )
        assert ra_list[0] == self.x_pos[0]
        assert len(ra_list) == 9

        ra_list, dec_list, amp_list = self.PointSource.point_source_list(
            self.kwargs_ps, self.kwargs_lens, k=0
        )
        assert ra_list[0] == self.x_pos[0]
        assert len(ra_list) == 4
        assert len(dec_list) == 4
        assert len(amp_list) == 4

    def test_point_source_amplitude(self):
        amp_list = self.PointSource.source_amplitude(self.kwargs_ps, self.kwargs_lens)
        assert len(amp_list) == 3

    def test_set_save_cache(self):
        self.PointSource.set_save_cache(True)
        assert self.PointSource._point_source_list[0]._save_cache == True

        self.PointSource.set_save_cache(False)
        assert self.PointSource._point_source_list[0]._save_cache == False

    def test_update_lens_model(self):
        lensModel = LensModel(lens_model_list=["SIS"])
        self.PointSource.update_lens_model(lens_model_class=lensModel)
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        x_image_list, y_image_list = self.PointSource.image_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=kwargs_lens
        )
        npt.assert_almost_equal(x_image_list[0][-1], -0.82654997748011705, decimal=8)

    def test_set_amplitudes(self):
        amp_list = [np.ones_like(self.x_pos) * 20, [100], np.ones_like(self.x_pos) * 10]
        kwargs_out = self.PointSource.set_amplitudes(amp_list, self.kwargs_ps)
        assert kwargs_out[0]["point_amp"][0] == 10 * self.kwargs_ps[0]["point_amp"][0]
        assert kwargs_out[1]["point_amp"][0] == 10 * self.kwargs_ps[1]["point_amp"][0]
        assert kwargs_out[2]["point_amp"][3] == 10 * self.kwargs_ps[2]["point_amp"][3]

    def test_update_search_window(self):
        search_window = 5
        x_center, y_center = 1, 1
        min_distance = 0.01

        point_source = PointSource(
            point_source_type_list=["LENSED_POSITION"],
            lensModel=None,
            kwargs_lens_eqn_solver={},
        )

        point_source.update_search_window(
            search_window,
            x_center,
            y_center,
            min_distance=min_distance,
            only_from_unspecified=False,
        )
        assert point_source._kwargs_lens_eqn_solver["search_window"] == search_window
        assert point_source._kwargs_lens_eqn_solver["x_center"] == x_center
        assert point_source._kwargs_lens_eqn_solver["x_center"] == y_center

        point_source = PointSource(
            point_source_type_list=["LENSED_POSITION"],
            lensModel=None,
            kwargs_lens_eqn_solver={},
        )

        point_source.update_search_window(
            search_window,
            x_center,
            y_center,
            min_distance=min_distance,
            only_from_unspecified=True,
        )
        assert point_source._kwargs_lens_eqn_solver["search_window"] == search_window
        assert point_source._kwargs_lens_eqn_solver["x_center"] == x_center
        assert point_source._kwargs_lens_eqn_solver["x_center"] == y_center

        kwargs_lens_eqn_solver = {
            "search_window": search_window,
            "min_distance": min_distance,
            "x_center": x_center,
            "y_center": y_center,
        }
        point_source = PointSource(
            point_source_type_list=["LENSED_POSITION"],
            lensModel=None,
            kwargs_lens_eqn_solver=kwargs_lens_eqn_solver,
        )
        point_source.update_search_window(
            search_window=-10,
            x_center=-10,
            y_center=-10,
            min_distance=10,
            only_from_unspecified=True,
        )
        assert point_source._kwargs_lens_eqn_solver["search_window"] == search_window
        assert point_source._kwargs_lens_eqn_solver["x_center"] == x_center
        assert point_source._kwargs_lens_eqn_solver["x_center"] == y_center

    def test__sort_position_by_original(self):
        from lenstronomy.PointSource.point_source import _sort_position_by_original

        x_o, y_o = np.array([1, 2]), np.array([0, 0])
        x_solved, y_solved = np.array([2]), np.array([0])
        x_new, y_new = _sort_position_by_original(x_o, y_o, x_solved, y_solved)
        npt.assert_almost_equal(x_new, x_o, decimal=7)
        npt.assert_almost_equal(y_new, y_o, decimal=7)

        x_solved, y_solved = np.array([2, 1]), np.array([0, 0.01])
        x_new, y_new = _sort_position_by_original(x_o, y_o, x_solved, y_solved)
        npt.assert_almost_equal(x_new, x_o, decimal=7)
        npt.assert_almost_equal(y_new, np.array([0.01, 0]), decimal=7)


class TestPointSourceFixedMag(object):
    def setup_method(self):
        lensModel = LensModel(lens_model_list=["SPEP"])
        solver = LensEquationSolver(lensModel=lensModel)
        e1, e2 = param_util.phi_q2_ellipticity(0, 0.7)
        self.kwargs_lens = [
            {
                "theta_E": 1.0,
                "center_x": 0,
                "center_y": 0,
                "e1": e1,
                "e2": e2,
                "gamma": 2,
            }
        ]
        self.sourcePos_x, self.sourcePos_y = 0.01, -0.01
        self.x_pos, self.y_pos = solver.image_position_from_source(
            sourcePos_x=self.sourcePos_x,
            sourcePos_y=self.sourcePos_y,
            kwargs_lens=self.kwargs_lens,
        )
        self.PointSource = PointSource(
            point_source_type_list=["LENSED_POSITION", "UNLENSED", "SOURCE_POSITION"],
            lensModel=lensModel,
            fixed_magnification_list=[True] * 4,
            additional_images_list=[False] * 4,
        )
        self.kwargs_ps = [
            {"ra_image": self.x_pos, "dec_image": self.y_pos, "source_amp": 1},
            {"ra_image": [1.0], "dec_image": [1.0], "point_amp": [10]},
            {
                "ra_source": self.sourcePos_x,
                "dec_source": self.sourcePos_y,
                "source_amp": 1.0,
            },
            {},
        ]

    def test_image_position(self):
        x_image_list, y_image_list = self.PointSource.image_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(x_image_list[0][0], self.x_pos[0], decimal=8)
        npt.assert_almost_equal(x_image_list[1], 1, decimal=8)
        npt.assert_almost_equal(x_image_list[2][0], self.x_pos[0], decimal=8)

    def test_source_position(self):
        x_source_list, y_source_list = self.PointSource.source_position(
            kwargs_ps=self.kwargs_ps, kwargs_lens=self.kwargs_lens
        )
        npt.assert_almost_equal(x_source_list[0], self.sourcePos_x, decimal=8)
        npt.assert_almost_equal(x_source_list[1], 1, decimal=8)
        npt.assert_almost_equal(x_source_list[2], self.sourcePos_x, decimal=8)

    def test_num_basis(self):
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert num_basis == 3

    def test_linear_response_set(self):
        ra_pos, dec_pos, amp, n = self.PointSource.linear_response_set(
            self.kwargs_ps, kwargs_lens=self.kwargs_lens, with_amp=False
        )
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert n == num_basis
        assert ra_pos[0][0] == self.x_pos[0]
        assert ra_pos[1][0] == 1
        assert np.all(amp != 1)
        npt.assert_almost_equal(ra_pos[2][0], self.x_pos[0], decimal=8)

        ra_pos, dec_pos, amp, n = self.PointSource.linear_response_set(
            self.kwargs_ps, kwargs_lens=self.kwargs_lens, with_amp=True
        )
        num_basis = self.PointSource.num_basis(self.kwargs_ps, self.kwargs_lens)
        assert n == num_basis
        assert ra_pos[0][0] == self.x_pos[0]
        assert ra_pos[1][0] == 1
        assert np.all(amp != 1)
        npt.assert_almost_equal(ra_pos[2][0], self.x_pos[0], decimal=8)

    def test_point_source_list(self):
        ra_list, dec_list, amp_list = self.PointSource.point_source_list(
            self.kwargs_ps, self.kwargs_lens
        )
        assert ra_list[0] == self.x_pos[0]
        assert len(ra_list) == 9

    def test_check_image_positions(self):
        bool = self.PointSource.check_image_positions(
            self.kwargs_ps, self.kwargs_lens, tolerance=0.001
        )
        assert bool is True

        # now we change the lens model to make the test fail
        kwargs_lens = [
            {"theta_E": 2.0, "center_x": 0, "center_y": 0, "e1": 0, "e2": 0, "gamma": 2}
        ]
        bool = self.PointSource.check_image_positions(
            self.kwargs_ps, kwargs_lens, tolerance=0.001
        )
        assert bool is False

    def test_set_amplitudes(self):
        amp_list = [10, [100], 10]
        kwargs_out = self.PointSource.set_amplitudes(amp_list, self.kwargs_ps)
        assert kwargs_out[0]["source_amp"] == 10 * self.kwargs_ps[0]["source_amp"]
        assert kwargs_out[1]["point_amp"][0] == 10 * self.kwargs_ps[1]["point_amp"][0]
        assert kwargs_out[2]["source_amp"] == 10 * self.kwargs_ps[2]["source_amp"]

    def test_positive_flux(self):
        bool = PointSource.check_positive_flux(
            kwargs_ps=[{"point_amp": np.array([1, -1])}]
        )
        assert bool is False
        bool = PointSource.check_positive_flux(kwargs_ps=[{"point_amp": -1}])
        assert bool is False

        bool = PointSource.check_positive_flux(
            kwargs_ps=[{"point_amp": np.array([0, 1])}]
        )
        assert bool is True
        bool = PointSource.check_positive_flux(kwargs_ps=[{"point_amp": 1}])
        assert bool is True

        bool = PointSource.check_positive_flux(
            kwargs_ps=[{"point_amp": np.array([0, 1]), "source_amp": 1}]
        )
        assert bool is True
        bool = PointSource.check_positive_flux(
            kwargs_ps=[{"point_amp": 1, "source_amp": -1}]
        )
        assert bool is False


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            PointSource(point_source_type_list=["BAD"])


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.Util.magnification_finite_util import (
    auto_raytracing_grid_resolution,
    auto_raytracing_grid_size,
)
import lenstronomy.Util.param_util as param_util
from lenstronomy.LightModel.light_model import LightModel
from astropy.cosmology import FlatLambdaCDM


class TestLensModelExtensions(object):
    """Tests the source model routines."""

    def setup_method(self):
        self.cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

    def test_critical_curves(self):
        lens_model_list = ["SPEP"]
        phi, q = 1.0, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.0,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        lens_model = LensModel(lens_model_list)
        lensModelExtensions = LensModelExtensions(LensModel(lens_model_list))

        # this first test is for even pixel grid numbers
        compute_window, grid_scale = 5, 0.05
        num_pix = int(compute_window / grid_scale)
        assert num_pix % 2 == 0
        (
            ra_crit_list,
            dec_crit_list,
            ra_caustic_list,
            dec_caustic_list,
        ) = lensModelExtensions.critical_curve_caustics(
            kwargs_lens, compute_window=compute_window, grid_scale=grid_scale
        )

        # here we test whether the caustic points are in fact at high magnifications (close to infinite)
        # close here means above magnification of 1000000 (with matplotlib method, this limit achieved was 170)
        for k in range(len(ra_crit_list)):
            ra_crit = ra_crit_list[k]
            dec_crit = dec_crit_list[k]
            mag = lens_model.magnification(ra_crit, dec_crit, kwargs_lens)
            assert np.all(np.abs(mag) > 1000)

        # this second test is for odd pixel grid numbers
        compute_window, grid_scale = 5.0 + 0.06, 0.05
        print(compute_window / grid_scale, "test float")
        num_pix = int(compute_window / grid_scale)
        print(num_pix, "test num_pix")
        assert num_pix % 2 == 1
        (
            ra_crit_list,
            dec_crit_list,
            ra_caustic_list,
            dec_caustic_list,
        ) = lensModelExtensions.critical_curve_caustics(
            kwargs_lens, compute_window=compute_window, grid_scale=grid_scale
        )

        # here we test whether the caustic points are in fact at high magnifications (close to infinite)
        # close here means above magnification of 1000000 (with matplotlib method, this limit achieved was 170)
        for k in range(len(ra_crit_list)):
            ra_crit = ra_crit_list[k]
            dec_crit = dec_crit_list[k]
            mag = lens_model.magnification(ra_crit, dec_crit, kwargs_lens)
            assert np.all(np.abs(mag) > 1000)

    def test_critical_curves_tiling(self):
        lens_model_list = ["SPEP"]
        phi, q = 1.0, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.0,
                "e1": e1,
                "e2": e2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        lensModel = LensModelExtensions(LensModel(lens_model_list))
        ra_crit, dec_crit = lensModel.critical_curve_tiling(
            kwargs_lens, compute_window=5, start_scale=0.01, max_order=10
        )
        # here we test whether the caustic points are in fact at high magnifications (close to infinite)
        # close here means above magnification of 1000. This is more precise than the critical_curve_caustics() method
        lens_model = LensModel(lens_model_list)
        mag = lens_model.magnification(ra_crit, dec_crit, kwargs_lens)
        assert np.all(np.abs(mag) > 1000)

    def test_get_magnification_model(self):
        self.kwargs_options = {
            "lens_model_list": ["GAUSSIAN"],
            "source_light_model_list": ["GAUSSIAN"],
            "lens_light_model_list": ["SERSIC"],
            "subgrid_res": 10,
            "numPix": 200,
            "psf_type": "gaussian",
            "x2_simple": True,
        }
        kwargs_lens = [
            {"amp": 1, "sigma_x": 2, "sigma_y": 2, "center_x": 0, "center_y": 0}
        ]

        x_pos = np.array([1.0, 1.0, 2.0])
        y_pos = np.array([-1.0, 0.0, 0.0])
        lens_model = LensModelExtensions(LensModel(lens_model_list=["GAUSSIAN"]))
        mag = lens_model.magnification_finite(
            x_pos,
            y_pos,
            kwargs_lens,
            source_sigma=0.003,
            window_size=0.1,
            grid_number=100,
        )
        npt.assert_almost_equal(mag[0], 0.98848384784633392, decimal=5)

    def test_magnification_finite(self):
        lens_model_list = ["SPEP", "SHEAR"]

        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.0,
                "e1": 0.02,
                "e2": -0.09,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.01, "gamma2": 0.03},
        ]

        extension = LensModelExtensions(LensModel(lens_model_list))
        x_image = [0.56153533, -0.78067875, -0.72551184, 0.75664112]
        y_image = [-0.74722528, 0.52491177, -0.72799235, 0.78503659]

        mag_square_grid = extension.magnification_finite(
            x_image,
            y_image,
            kwargs_lens,
            source_sigma=0.001,
            grid_number=200,
            window_size=0.1,
        )
        mag_polar_grid = extension.magnification_finite(
            x_image,
            y_image,
            kwargs_lens,
            source_sigma=0.001,
            grid_number=200,
            window_size=0.1,
            polar_grid=True,
        )
        npt.assert_almost_equal(mag_polar_grid, mag_square_grid, decimal=5)

    def test_magnification_finite_adaptive(self):
        lens_model_list = ["EPL", "SHEAR"]
        z_source = 1.5
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": 2.0,
                "e1": 0.02,
                "e2": -0.09,
                "center_x": 0,
                "center_y": 0,
            },
            {"gamma1": 0.01, "gamma2": 0.03},
        ]

        lensmodel = LensModel(lens_model_list)
        extension = LensModelExtensions(lensmodel)
        solver = LensEquationSolver(lensmodel)
        source_x, source_y = 0.07, 0.03
        x_image, y_image = solver.findBrightImage(source_x, source_y, kwargs_lens)

        source_fwhm_parsec = 40.0

        pc_per_arcsec = 1000 / self.cosmo.arcsec_per_kpc_proper(z_source).value
        source_sigma = source_fwhm_parsec / pc_per_arcsec / 2.355

        grid_size = auto_raytracing_grid_size(source_fwhm_parsec)
        grid_resolution = auto_raytracing_grid_resolution(source_fwhm_parsec)
        # make this even higher resolution to show convergence
        grid_number = int(2 * grid_size / grid_resolution)
        window_size = 2 * grid_size

        mag_square_grid = extension.magnification_finite(
            x_image,
            y_image,
            kwargs_lens,
            source_sigma=source_sigma,
            grid_number=grid_number,
            window_size=window_size,
        )
        flux_ratios_square_grid = mag_square_grid / max(mag_square_grid)

        mag_adaptive_grid = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=self.cosmo,
        )
        flux_ratios_adaptive_grid = mag_adaptive_grid / max(mag_adaptive_grid)

        mag_adaptive_grid_fixed_aperture_size = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=self.cosmo,
            fixed_aperture_size=True,
            grid_radius_arcsec=0.2,
        )
        flux_ratios_fixed_aperture_size = mag_adaptive_grid_fixed_aperture_size / max(
            mag_adaptive_grid_fixed_aperture_size
        )

        mag_adaptive_grid_2 = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=self.cosmo,
            axis_ratio=0,
        )
        mag_adaptive_grid_3 = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=self.cosmo,
            axis_ratio=1,
        )

        flux_ratios_adaptive_grid_2 = mag_adaptive_grid_2 / max(mag_adaptive_grid_2)
        flux_ratios_adaptive_grid_3 = mag_adaptive_grid_3 / max(mag_adaptive_grid_3)

        # tests the default cosmology
        _ = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=None,
            tol=0.0001,
        )

        # test smallest eigenvalue
        _ = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=None,
            tol=0.0001,
            use_largest_eigenvalue=False,
        )

        # tests the r_max > sqrt(2) * grid_radius stop criterion
        _ = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=None,
            tol=0.0001,
            step_size=1000,
        )

        mag_point_source = abs(lensmodel.magnification(x_image, y_image, kwargs_lens))

        quarter_precent_precision = [0.0025] * 4
        npt.assert_array_less(
            flux_ratios_square_grid / flux_ratios_adaptive_grid - 1,
            quarter_precent_precision,
        )
        npt.assert_array_less(
            flux_ratios_fixed_aperture_size / flux_ratios_adaptive_grid - 1,
            quarter_precent_precision,
        )
        npt.assert_array_less(
            flux_ratios_square_grid / flux_ratios_adaptive_grid_2 - 1,
            quarter_precent_precision,
        )
        npt.assert_array_less(
            flux_ratios_adaptive_grid_3 / flux_ratios_adaptive_grid_2 - 1,
            quarter_precent_precision,
        )
        half_percent_precision = [0.005] * 4
        npt.assert_array_less(
            mag_square_grid / mag_adaptive_grid - 1, half_percent_precision
        )
        npt.assert_array_less(
            mag_square_grid / mag_adaptive_grid_2 - 1, half_percent_precision
        )
        npt.assert_array_less(
            mag_adaptive_grid / mag_point_source - 1, half_percent_precision
        )

        flux_array = np.array([0.0, 0.0])
        grid_x = np.array([0.0, source_sigma])
        grid_y = np.array([0.0, 0.0])
        grid_r = np.hypot(grid_x, grid_y)

        # test that the double gaussian profile has 2x the flux when dx, dy = 0
        magnification_double_gaussian = extension.magnification_finite_adaptive(
            x_image,
            y_image,
            source_x,
            source_y,
            kwargs_lens,
            source_fwhm_parsec,
            z_source,
            cosmo=self.cosmo,
            source_light_model="DOUBLE_GAUSSIAN",
            dx=0.0,
            dy=0.0,
            amp_scale=1.0,
            size_scale=1.0,
        )
        npt.assert_almost_equal(magnification_double_gaussian, 2 * mag_adaptive_grid)

        grid_radius = 0.3
        npix = 400
        _x = _y = np.linspace(-grid_radius, grid_radius, npix)
        resolution = 2 * grid_radius / npix
        xx, yy = np.meshgrid(_x, _y)
        for i in range(0, 4):
            beta_x, beta_y = lensmodel.ray_shooting(
                x_image[i] + xx.ravel(), y_image[i] + yy.ravel(), kwargs_lens
            )
            source_light_model = LightModel(["GAUSSIAN"] * 2)
            amp_scale, dx, dy, size_scale = 0.2, 0.005, -0.005, 1.0
            kwargs_source = [
                {
                    "amp": 1.0,
                    "center_x": source_x,
                    "center_y": source_y,
                    "sigma": source_sigma,
                },
                {
                    "amp": amp_scale,
                    "center_x": source_x + dx,
                    "center_y": source_y + dy,
                    "sigma": source_sigma * size_scale,
                },
            ]
            sb = source_light_model.surface_brightness(beta_x, beta_y, kwargs_source)
            magnification_double_gaussian_reference = np.sum(sb) * resolution**2
            magnification_double_gaussian = extension.magnification_finite_adaptive(
                [x_image[i]],
                [y_image[i]],
                source_x,
                source_y,
                kwargs_lens,
                source_fwhm_parsec,
                z_source,
                cosmo=self.cosmo,
                source_light_model="DOUBLE_GAUSSIAN",
                dx=dx,
                dy=dy,
                amp_scale=amp_scale,
                size_scale=size_scale,
                grid_resolution=resolution,
                grid_radius_arcsec=grid_radius,
                axis_ratio=1.0,
            )
            npt.assert_almost_equal(
                magnification_double_gaussian / magnification_double_gaussian_reference,
                1.0,
                3,
            )

        source_model = LightModel(["GAUSSIAN"])
        kwargs_source = [
            {
                "amp": 1.0,
                "center_x": source_x,
                "center_y": source_y,
                "sigma": source_sigma,
            }
        ]

        r_min = 0.0
        r_max = source_sigma * 0.9
        flux_array = extension._magnification_adaptive_iteration(
            flux_array,
            x_image[0],
            y_image[0],
            grid_x,
            grid_y,
            grid_r,
            r_min,
            r_max,
            lensmodel,
            kwargs_lens,
            source_model,
            kwargs_source,
        )
        bx, by = lensmodel.ray_shooting(x_image[0], y_image[0], kwargs_lens)
        sb_true = source_model.surface_brightness(bx, by, kwargs_source)
        npt.assert_equal(True, flux_array[0] == sb_true)
        npt.assert_equal(True, flux_array[1] == 0.0)

        r_min = source_sigma * 0.9
        r_max = 2 * source_sigma

        flux_array = extension._magnification_adaptive_iteration(
            flux_array,
            x_image[0],
            y_image[0],
            grid_x,
            grid_y,
            grid_r,
            r_min,
            r_max,
            lensmodel,
            kwargs_lens,
            source_model,
            kwargs_source,
        )
        bx, by = lensmodel.ray_shooting(
            x_image[0] + source_sigma, y_image[0], kwargs_lens
        )
        sb_true = source_model.surface_brightness(bx, by, kwargs_source)
        npt.assert_equal(True, flux_array[1] == sb_true)

    def test_zoom_source(self):
        lens_model_list = ["SIE", "SHEAR"]
        lensModel = LensModel(lens_model_list=lens_model_list)
        lensModelExtensions = LensModelExtensions(lensModel=lensModel)
        lensEquationSolver = LensEquationSolver(lensModel=lensModel)

        x_source, y_source = 0.02, 0.01
        kwargs_lens = [
            {"theta_E": 1, "e1": 0.1, "e2": 0.1, "center_x": 0, "center_y": 0},
            {"gamma1": 0.05, "gamma2": -0.03},
        ]

        x_img, y_img = lensEquationSolver.image_position_from_source(
            kwargs_lens=kwargs_lens, sourcePos_x=x_source, sourcePos_y=y_source
        )

        image = lensModelExtensions.zoom_source(
            x_img[0],
            y_img[0],
            kwargs_lens,
            source_sigma=0.003,
            window_size=0.1,
            grid_number=100,
            shape="GAUSSIAN",
        )
        assert len(image) == 100

    def test_tangential_average(self):
        lens_model_list = ["SIS"]
        lensModel = LensModel(lens_model_list=lens_model_list)
        lensModelExtensions = LensModelExtensions(lensModel=lensModel)
        tang_stretch_ave = lensModelExtensions.tangential_average(
            x=1.1,
            y=0,
            kwargs_lens=[{"theta_E": 1, "center_x": 0, "center_y": 0}],
            dr=1,
            smoothing=None,
            num_average=9,
        )
        npt.assert_almost_equal(tang_stretch_ave, -2.525501464097973, decimal=6)

    def test_caustic_area(self):
        lens_model_list = ["SIE"]
        lensModel = LensModel(lens_model_list=lens_model_list)
        lensModelExtensions = LensModelExtensions(lensModel=lensModel)
        kwargs_lens = [{"theta_E": 1, "e1": 0.2, "e2": 0, "center_x": 0, "center_y": 0}]
        kwargs_caustic_num = {
            "compute_window": 3,
            "grid_scale": 0.005,
            "center_x": 0,
            "center_y": 0,
        }

        area = lensModelExtensions.caustic_area(
            kwargs_lens=kwargs_lens,
            kwargs_caustic_num=kwargs_caustic_num,
            index_vertices=0,
        )
        npt.assert_almost_equal(area, 0.08445866728739478, decimal=3)



__author__ = "aymgal"

import numpy as np
from scipy import stats

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


# transform the unit hypercube to pysical parameters for (nested) sampling


@export
def unit2uniform(x, vmin, vmax):
    """Mapping from uniform distribution on parameter space to uniform distribution on
    unit hypercube."""
    return vmin + (vmax - vmin) * x


@export
def uniform2unit(theta, vmin, vmax):
    """Mapping from uniform distribution on unit hypercube to uniform distribution on
    parameter space."""
    return (theta - vmin) / (vmax - vmin)


@export
def cube2args_uniform(cube, lowers, uppers, num_dims, copy=False):
    """Mapping from uniform distribution on unit hypercube 'cube' to uniform
    distribution on parameter space.

    :param cube: list or 1D-array of parameter values on unit hypercube
    :param lowers: lower bounds for each parameter
    :param uppers: upper bounds for each parameter
    :param num_dims: parameter space dimension (= number of parameters)
    :param copy: If False, this function modifies 'cube' in-place. Default to False.
    :return: hypercube mapped to parameters space
    """
    if copy:
        cube_ = cube
        cube = np.zeros_like(cube_)
    for i in range(num_dims):
        val = cube_[i] if copy else cube[i]
        low, upp = lowers[i], uppers[i]
        cube[i] = unit2uniform(val, low, upp)
    return cube


@export
def cube2args_gaussian(cube, lowers, uppers, means, sigmas, num_dims, copy=False):
    """Mapping from uniform distribution on unit hypercube 'cube' to truncated gaussian
    distribution on parameter space, with mean 'mu' and std dev 'sigma'.

    :param cube: list or 1D-array of parameter values on unit hypercube
    :param lowers: lower bounds for each parameter
    :param uppers: upper bounds for each parameter
    :param means: gaussian mean for each parameter
    :param sigmas: gaussian std deviation for each parameter
    :param num_dims: parameter space dimension (= number of parameters)
    :param copy: If False, this function modifies 'cube' in-place. Default to False.
    :return: hypercube mapped to parameters space
    """
    if copy:
        cube_ = cube
        cube = np.zeros_like(cube_)
    a, b = (np.array(lowers) - means) / sigmas, (np.array(uppers) - means) / sigmas
    cube[:] = stats.truncnorm.ppf(
        cube_ if copy else cube, a=a, b=b, loc=means, scale=sigmas
    )
    return cube


@export
def scale_limits(lowers, uppers, scale):
    if not isinstance(lowers, np.ndarray):
        lowers = np.asarray(lowers)
        uppers = np.asarray(uppers)
    mid_points = (lowers + uppers) / 2.0
    widths_scaled = (uppers - lowers) * scale
    lowers_scaled = mid_points - widths_scaled / 2.0
    uppers_scaled = mid_points + widths_scaled / 2.0
    return lowers_scaled, uppers_scaled


@export
def sample_ball(p0, std, size=1, dist="uniform"):
    """Produce a ball of walkers around an initial parameter value. this routine is from
    the emcee package as it became deprecated there.

    :param p0: The initial parameter values (array).
    :param std: The axis-aligned standard deviation (array).
    :param size: The number of samples to produce.
    :param dist: string, specifies the distribution being sampled, supports 'uniform'
        and 'normal'
    """
    assert len(p0) == len(std)
    if dist == "uniform":
        return np.vstack(
            [
                p0 + std * np.random.uniform(low=-1, high=1, size=len(p0))
                for i in range(size)
            ]
        )
    elif dist == "normal":
        return np.vstack(
            [
                p0 + std * np.random.normal(loc=0, scale=1, size=len(p0))
                for i in range(size)
            ]
        )
    else:
        raise ValueError(
            'distribution %s not supported. Chose among "uniform" or "normal".' % dist
        )


@export
def sample_ball_truncated(mean, sigma, lower_limit, upper_limit, size):
    """Samples gaussian ball with truncation at lower and upper limit of the
    distribution.

    :param mean: numpy array, mean of the distribution to be sampled
    :param sigma: numpy array, sigma of the distribution to be sampled
    :param lower_limit: numpy array, lower bound of to be sampled distribution
    :param upper_limit: numpy array, upper bound of to be sampled distribution
    :param size: number of tuples to be sampled
    :return: realization of truncated normal distribution with shape (size,
        dim(parameters))
    """
    a, b = (lower_limit - mean) / sigma, (upper_limit - mean) / sigma
    draws = np.vstack(
        [mean + sigma * stats.truncnorm.rvs(a, b, size=len(a)) for i in range(size)]
    )
    return draws



__author__ = "sibirrer"

import pytest
import numpy.testing as npt
import numpy as np
from lenstronomy.Sampling.special_param import SpecialParam


class TestParam(object):
    def setup_method(self):
        self.param = SpecialParam(
            Ddt_sampling=True,
            kwargs_fixed=None,
            point_source_offset=True,
            num_images=2,
            source_size=True,
            num_tau0=2,
            num_z_sampling=3,
            source_grid_offset=True,
            kinematic_sampling=True,
            kwargs_lower={"z_sampling": [0.05, 0.1, 0.5]},
            kwargs_upper={"z_sampling": [0.2, 1.0, 1.0]},
        )
        self.kwargs = {
            "D_dt": 1988,
            "delta_x_image": [0, 0],
            "delta_y_image": [0, 0],
            "source_size": 0.1,
            "tau0_list": [0, 1],
            "z_sampling": np.array([0.1, 0.5, 2]),
            "delta_x_source_grid": 0,
            "delta_y_source_grid": 0,
            "b_ani": 0.1,
            "incli": 0.0,
            "D_d": 2000,
        }

    def test_get_setParams(self):
        args = self.param.set_params(self.kwargs)
        kwargs_new, _ = self.param.get_params(args, i=0)
        args_new = self.param.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

        param_fixed = SpecialParam(
            Ddt_sampling=True,
            kwargs_fixed=self.kwargs,
            point_source_offset=True,
            num_images=2,
            source_size=True,
            num_z_sampling=3,
            num_tau0=2,
            kinematic_sampling=True,
        )
        kwargs_new, i = param_fixed.get_params(args=[], i=0)
        kwargs_new["D_dt"] = self.kwargs["D_dt"]
        kwargs_new["b_ani"] = self.kwargs["b_ani"]
        kwargs_new["incli"] = self.kwargs["incli"]
        kwargs_new["D_d"] = self.kwargs["D_d"]

        special_param = SpecialParam(
            num_z_sampling=1,
            kwargs_lower={"z_sampling": [0.1]},
            kwargs_upper={"z_sampling": [0.2]},
        )

        kwargs_test, i = special_param.get_params(args=[0.3], i=0, impose_bound=True)
        assert kwargs_test["z_sampling"] == [0.2]

    def test_num_params(self):
        num, list = self.param.num_param()
        assert num == 16

    def test_mass_scaling(self):
        kwargs_fixed = {}
        param = SpecialParam(
            kwargs_fixed=kwargs_fixed, mass_scaling=True, num_scale_factor=3
        )
        kwargs = {"scale_factor": [0, 1, 2]}
        args = param.set_params(kwargs)
        assert len(args) == 3
        num_param, param_list = param.num_param()
        assert num_param == 3
        kwargs_new, _ = param.get_params(args, i=0)
        assert kwargs_new["scale_factor"][1] == 1
        param = SpecialParam(kwargs_fixed=kwargs, mass_scaling=True, num_scale_factor=3)
        kwargs_in = {"scale_factor": [9, 9, 9]}
        args = param.set_params(kwargs_in)
        assert len(args) == 0
        kwargs_new, _ = param.get_params(args, i=0)
        print(kwargs_new)
        assert kwargs_new["scale_factor"][1] == 1

    def test_delta_images(self):
        param = SpecialParam(
            num_images=2,
            point_source_offset=True,
            kwargs_fixed={},
            kwargs_lower={"delta_x_image": [-1, -1], "delta_y_image": [-1, -1]},
            kwargs_upper={"delta_x_image": [1, 1], "delta_y_image": [1, 1]},
        )
        kwargs = {"delta_x_image": [0.5, 0.5], "delta_y_image": [0.5, 0.5]}
        args = param.set_params(kwargs_special=kwargs)
        kwargs_new, _ = param.get_params(args, i=0)
        print(kwargs_new)
        assert kwargs_new["delta_x_image"][0] == kwargs["delta_x_image"][0]

    def test_source_grid_offsets(self):
        param = SpecialParam(
            kwargs_lower={"delta_x_source_grid": -1, "delta_y_source_grid": 1},
            kwargs_upper={"delta_x_source_grid": 1, "delta_y_source_grid": 1},
            source_grid_offset=True,
        )
        kwargs = {"delta_x_source_grid": 0.1, "delta_y_source_grid": 0.1}
        args = param.set_params(kwargs_special=kwargs)
        kwargs_new, _ = param.get_params(args, i=0)
        assert kwargs_new["delta_x_source_grid"] == kwargs["delta_x_source_grid"]
        assert kwargs_new["delta_y_source_grid"] == kwargs["delta_y_source_grid"]

        kwargs_fixed = {"delta_x_source_grid": 0, "delta_y_source_grid": 0}
        param = SpecialParam(
            kwargs_lower={"delta_x_source_grid": -1, "delta_y_source_grid": 1},
            kwargs_upper={"delta_x_source_grid": 1, "delta_y_source_grid": 1},
            source_grid_offset=True,
            kwargs_fixed=kwargs_fixed,
        )
        kwargs = {"delta_x_source_grid": 0.1, "delta_y_source_grid": 0.1}
        args = param.set_params(kwargs_special=kwargs)
        kwargs_new, _ = param.get_params(args, i=0)
        assert kwargs_new["delta_x_source_grid"] == kwargs_fixed["delta_x_source_grid"]
        assert kwargs_new["delta_y_source_grid"] == kwargs_fixed["delta_y_source_grid"]

    def test_general_scaling(self):
        kwargs_fixed = {}
        param = SpecialParam(
            kwargs_fixed=kwargs_fixed,
            general_scaling_params={"param": [False, 1, 1, False, 2]},
        )
        args = param.set_params(
            {"param_scale_factor": [1, 2], "param_scale_pow": [3, 4]}
        )
        assert len(args) == 4
        num_param, param_list = param.num_param()
        assert num_param == 4
        kwargs_new, _ = param.get_params(args, i=0)
        assert kwargs_new["param_scale_factor"] == [1, 2]
        assert kwargs_new["param_scale_pow"] == [3, 4]


if __name__ == "__main__":
    pytest.main()



__author__ = "pierrefleury"

__all__ = ["LOS"]


class LOS(object):
    """Class allowing one to add tidal line-of-sight effects (convergence and shear) to
    single-plane lensing. Stricly speaking, this is not a profile, but when present in
    list of lens models, it is automatically recognised by ModelAPI(), which sets the
    flag los_effects to True, and thereby leads LensModel to use SinglePlaneLOS()
    instead of SinglePlane(). It is however incompatible with MultiPlane().

    The key-word arguments are the three line-of-sight convergences, the two components
    of the three line-of-sight shears, and the three line-of-sight rotations, all
    defined with the convention of
    https://arxiv.org/abs/2104.08883:
    kappa_od, kappa_os, kappa_ds, gamma1_od, gamma2_od, gamma1_os, gamma2_os,
    gamma1_ds, gamma2_ds, omega_od, omega_os, omega_ds

    Because LOS is not a profile, it does not contain the usual functions
    function(), derivatives(), and hessian(), but rather modifies the
    behaviour of those functions in the SinglePlaneLOS() class.

    Instead, it contains the essential building blocks of this modification.
    """

    param_names = [
        "kappa_od",
        "kappa_os",
        "kappa_ds",
        "gamma1_od",
        "gamma2_od",
        "gamma1_os",
        "gamma2_os",
        "gamma1_ds",
        "gamma2_ds",
        "omega_od",
        "omega_os",
        "omega_ds",
    ]
    lower_limit_default = {pert: -0.5 for pert in param_names}
    upper_limit_default = {pert: 0.5 for pert in param_names}

    def __init__(self, *args, **kwargs):
        self._static = False

    @staticmethod
    def distort_vector(x, y, kappa=0, gamma1=0, gamma2=0, omega=0):
        """This function applies a distortion matrix to a vector (x, y) and returns (x',
        y') as follows:

        .. math::
            \\begin{pmatrix}
            x'

            y'
            \\end{pmatrix}
            =
            \\begin{pmatrix}
            1 - \\kappa - \\gamma_1 & -\\gamma_2 + \\omega

            -\\gamma_2 - \\omega & 1 - \\kappa + \\gamma_1
            \\end{pmatrix}
            \\begin{pmatrix}
            x

            y
            \\end{pmatrix}

        :param x: x-component of the vector to which the distortion matrix is applied
        :param y: y-component of the vector to which the distortion matrix is applied
        :param kappa: the convergence
        :param gamma1: the first shear component
        :param gamma2: the second shear component
        :param omega: the rotation
        :return: the distorted vector
        """

        x_ = (1 - kappa - gamma1) * x + (-gamma2 + omega) * y
        y_ = (1 - kappa + gamma1) * y - (gamma2 + omega) * x

        return x_, y_

    @staticmethod
    def left_multiply(f_xx, f_xy, f_yx, f_yy, kappa=0, gamma1=0, gamma2=0, omega=0):
        """Left-multiplies the Hessian matrix of a lens with a distortion matrix with
        convergence kappa, shear gamma1, gamma2, and rotation omega:

        .. math::
            \\mathsf{H}'
            =
            \\begin{pmatrix}
            1 - \\kappa - \\gamma_1 & -\\gamma_2 + \\omega

            -\\gamma_2 - \\omega & 1 - \\kappa + \\gamma_1
            \\end{pmatrix}
            \\mathsf{H}

        :param f_xx: the i, i element of the Hessian matrix
        :param f_xy: the i, j element of the Hessian matrix
        :param f_yx: the j, i element of the Hessian matrix
        :param f_yy: the j, j element of the Hessian matrix
        :param kappa: the convergence
        :param gamma1: the first shear component
        :param gamma2: the second shear component
        :param omega: the rotation
        :return: the Hessian left-multiplied by the distortion matrix
        """

        f__xx = (1 - kappa - gamma1) * f_xx + (-gamma2 + omega) * f_yx
        f__xy = (1 - kappa - gamma1) * f_xy + (-gamma2 + omega) * f_yy
        f__yx = -(gamma2 + omega) * f_xx + (1 - kappa + gamma1) * f_yx
        f__yy = -(gamma2 + omega) * f_xy + (1 - kappa + gamma1) * f_yy

        return f__xx, f__xy, f__yx, f__yy

    @staticmethod
    def right_multiply(f_xx, f_xy, f_yx, f_yy, kappa=0, gamma1=0, gamma2=0, omega=0):
        """Right-multiplies the Hessian matrix of a lens with a distortion matrix with
        convergence kappa and shear gamma1, gamma2:

        .. math::
            \\mathsf{H}'
            =
            \\mathsf{H}
            \\begin{pmatrix}
            1 - \\kappa - \\gamma_1 & -\\gamma_2 + \\omega

            -\\gamma_2 - \\omega & 1 - \\kappa + \\gamma_1
            \\end{pmatrix}

        :param f_xx: the i, i element of the Hessian matrix
        :param f_xy: the i, j element of the Hessian matrix
        :param f_yx: the j, i element of the Hessian matrix
        :param f_yy: the j, j element of the Hessian matrix
        :param kappa: the convergence
        :param gamma1: the first shear component
        :param gamma2: the second shear component
        :param omega: the rotation
        :return: the Hessian right-multiplied by the distortion matrix
        """

        f__xx = (1 - kappa - gamma1) * f_xx - (gamma2 + omega) * f_xy
        f__xy = (-gamma2 + omega) * f_xx + (1 - kappa + gamma1) * f_xy
        f__yx = (1 - kappa - gamma1) * f_yx - (gamma2 + omega) * f_yy
        f__yy = (-gamma2 + omega) * f_yx + (1 - kappa + gamma1) * f_yy

        return f__xx, f__xy, f__yx, f__yy

    def set_static(self, **kwargs):
        """Pre-computes certain computations that do only relate to the lens model
        parameters and not to the specific position where to evaluate the lens model.

        :param kwargs: lens model parameters
        :return: no return, for certain lens model some private self variables are
            initiated
        """
        pass

    def set_dynamic(self):
        """

        :return: no return, deletes pre-computed variables for certain lens models
        """
        pass



import numpy as np

from lenstronomy.Util.numba_util import jit
from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def cart2polar(x, y, center_x=0, center_y=0):
    """Transforms cartesian coords [x,y] into polar coords [r,phi] in the frame of the
    lens center.

    :param x: set of x-coordinates
    :type x: array of size (n)
    :param y: set of x-coordinates
    :type y: array of size (n)
    :param center_x: rotation point
    :type center_x: float
    :param center_y: rotation point
    :type center_y: float
    :returns: array of same size with coords [r,phi]
    """
    coord_shift_x = x - center_x
    coord_shift_y = y - center_y
    r = np.sqrt(coord_shift_x**2 + coord_shift_y**2)
    phi = np.arctan2(coord_shift_y, coord_shift_x)
    return r, phi


@export
def polar2cart(r, phi, center):
    """Transforms polar coords [r,phi] into cartesian coords [x,y] in the frame of the
    lense center.

    :param r: radial coordinate (distance) to the center
    :type r: array of size n or float
    :param phi: angular coordinate
    :type phi: array of size n or float
    :param center: rotation point
    :type center: array of size (2)
    :returns: array of same size with coords [x,y]
    :raises: AttributeError, KeyError
    """
    x = r * np.cos(phi)
    y = r * np.sin(phi)
    return x - center[0], y - center[1]


@export
def shear_polar2cartesian(phi, gamma):
    """

    :param phi: shear angle (radian)
    :param gamma: shear strength
    :return: shear components gamma1, gamma2
    """
    gamma1 = gamma * np.cos(2 * phi)
    gamma2 = gamma * np.sin(2 * phi)
    return gamma1, gamma2


@export
def shear_cartesian2polar(gamma1, gamma2):
    """
    :param gamma1: cartesian shear component
    :param gamma2: cartesian shear component
    :return: shear angle, shear strength
    """
    phi = np.arctan2(gamma2, gamma1) / 2
    gamma = np.sqrt(gamma1**2 + gamma2**2)
    return phi, gamma


@export
@jit()
def phi_q2_ellipticity(phi, q):
    """Transforms orientation angle and axis ratio into complex ellipticity moduli e1,
    e2.

    :param phi: angle of orientation (in radian)
    :param q: axis ratio minor axis / major axis
    :return: eccentricities e1 and e2 in complex ellipticity moduli
    """
    e1 = (1.0 - q) / (1.0 + q) * np.cos(2 * phi)
    e2 = (1.0 - q) / (1.0 + q) * np.sin(2 * phi)
    return e1, e2


@export
@jit()
def ellipticity2phi_q(e1, e2):
    """Transforms complex ellipticity moduli in orientation angle and axis ratio.

    :param e1: eccentricity in x-direction
    :param e2: eccentricity in xy-direction
    :return: angle in radian, axis ratio (minor/major)
    """
    phi = np.arctan2(e2, e1) / 2
    c = np.sqrt(e1**2 + e2**2)
    c = np.minimum(c, 0.9999)
    q = (1 - c) / (1 + c)
    return phi, q


@export
def transform_e1e2_product_average(x, y, e1, e2, center_x, center_y):
    """Maps the coordinates x, y with eccentricities e1 e2 into a new elliptical
    coordinate system such that R = sqrt(R_major * R_minor)

    :param x: x-coordinate
    :param y: y-coordinate
    :param e1: eccentricity
    :param e2: eccentricity
    :param center_x: center of distortion
    :param center_y: center of distortion
    :return: distorted coordinates x', y'
    """
    phi_g, q = ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y

    cos_phi = np.cos(phi_g)
    sin_phi = np.sin(phi_g)

    xt1 = cos_phi * x_shift + sin_phi * y_shift
    xt2 = -sin_phi * x_shift + cos_phi * y_shift
    return xt1 * np.sqrt(q), xt2 / np.sqrt(q)


@export
def transform_e1e2_square_average(x, y, e1, e2, center_x, center_y):
    """Maps the coordinates x, y with eccentricities e1 e2 into a new elliptical
    coordinate system such that R = sqrt(R_major**2 + R_minor**2)

    :param x: x-coordinate
    :param y: y-coordinate
    :param e1: eccentricity
    :param e2: eccentricity
    :param center_x: center of distortion
    :param center_y: center of distortion
    :return: distorted coordinates x', y'
    """
    phi_g, q = ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y
    cos_phi = np.cos(phi_g)
    sin_phi = np.sin(phi_g)
    e = q2e(q)
    x_ = (cos_phi * x_shift + sin_phi * y_shift) * np.sqrt(1 - e)
    y_ = (-sin_phi * x_shift + cos_phi * y_shift) * np.sqrt(1 + e)
    return x_, y_


def q2e(q):
    """computes.

    .. math::
        e = \\equic \\frac{1 - q^2}{1 + q^2}

    :param q: axis ratio of minor to major axis
    :return: ellipticity e
    """
    e = abs(1 - q**2) / (1 + q**2)
    return e
